import numpy as np

from smrf.envphys.constants import EMISS_VEG, FREEZE, STEF_BOLTZ


def thermal_correct_canopy(th, ta, tau, veg_height, height_thresh=2):
    """
    Correct thermal radiation for vegitation.  It will only correct
    for pixels where the veg height is above a threshold. This ensures
    that the open areas don't get this applied.  Vegitation temp
    is assumed to be at air temperature

    Args:
        th: thermal radiation
        ta: air temperature [C]
        tau: transmissivity of the canopy
        veg_height: vegitation height for each pixel
        height_thresh: threshold hold for height to say that there is veg in
            the pixel

    Returns:
        corrected thermal radiation

    Equations from Link and Marks 1999 :cite:`Link&Marks:1999`

    20150611 Scott Havens
    """

    # thermal emitted from the canopy
    veg = STEF_BOLTZ * EMISS_VEG * np.power(ta + FREEZE, 4)

    # pixels with canopy above the threshold
    ind = veg_height > height_thresh

    # correct incoming thermal
    th[ind] = tau[ind] * th[ind] + (1 - tau[ind]) * veg[ind]

    return th



import logging

from topocalc.shade import shade

from smrf.envphys import sunang
from smrf.envphys.solar.irradiance import direct_solar_irradiance
from smrf.envphys.solar.twostream import twostream


def shade_thread(queue, date, sin_slope, aspect, zenith=None):
    """
    See shade for input argument descriptions

    Args:
        queue: queue with illum_ang, cosz, azimuth
        date_time: loop through dates to accesss
        sin_slope: numpy array of sine of slope angles sin(S)
        aspect: numpy array of aspect in radians from south
        azimuth: azimuth in degrees to the sun -180..180 (comes from sunang)
        zenith: the solar zenith angle 0..90 degrees

    """

    for v in ['cosz', 'azimuth', 'illum_ang']:
        if v not in queue.keys():
            raise ValueError('Queue must have {} key'.format(v))

    log = logging.getLogger(__name__)

    for t in date:

        log.debug('%s Calculating illumination angle' % t)

        mu = None
        cosz = queue['cosz'].get(t)

        if cosz > 0:
            azimuth = queue['azimuth'].get(t)
            mu = shade(sin_slope, aspect, azimuth, cosz, zenith)

        queue['illum_ang'].put([t, mu])


def model_solar(dt, lat, lon, tau=0.2, tzone=0):
    """
    Model solar radiation at a point
    Combines sun angle, solar and two stream

    Args:
        dt - datetime object
        lat - latitude
        lon - longitude
        tau - optical depth
        tzone - time zone

    Returns:
        corrected solar radiation
    """

    # determine the sun angle
    cosz, az, rad_vec = sunang.sunang(dt, lat, lon)

    # calculate the solar irradiance
    sol = direct_solar_irradiance(dt, [0.28, 2.8])

    # calculate the two stream model value
    R = twostream(cosz, sol, tau=tau)

    return R['irradiance_at_bottom']




from inicheck.tools import cast_all_variables

from smrf.framework.model_framework import SMRF, run_smrf
from smrf.tests.smrf_test_case import SMRFTestCase


class TestLoadCSVData(SMRFTestCase):

    def test_station_dates(self):
        """
        Test the start date not in the data
        """
        config = self.base_config

        # Use dates not in the dataset, expecting an error
        config.raw_cfg['time']['start_date'] = '1900-01-01 00:00'
        config.raw_cfg['time']['end_date'] = '1900-02-01 00:00'

        # apply the new recipes
        config.apply_recipes()
        config = cast_all_variables(config, config.mcfg)

        with self.assertRaises(Exception):
            run_smrf(config)

    def test_all_stations(self):
        """
        Test using all stations
        """

        # test the end date
        config = self.base_config
        config.raw_cfg['csv']['stations'] = ['RMESP', 'RME_176']

        # apply the new recipies
        config.apply_recipes()
        config = cast_all_variables(config, config.mcfg)

        self.assertIsInstance(run_smrf(config), SMRF)



from .utils import calc_percent_snow, check_temperature


class PiecewiseSusong1999:
    """
    Follows :func:`~smrf.envphys.snow.susong1999` but is the piecewise form
    of table shown there. This model adds to the former by accounting for
    liquid water effect near 0.0 Degrees C.

    The table was estimated by Danny Marks in 2017 which resulted in the
    piecewise equations below:

        Percent Snow:
            .. math::

                \\%_{snow} = \\Bigg \\lbrace{
                    \\frac{-T}{T_{r0}} P_{cr0} + P_{c0}, \\hfill
                    -0.5^{\\circ} C \\leq T \\leq 0.0^{\\circ} C
                     \\atop
                     \\frac{-T_{pp}}{T_{max} + 1.0} P_{c0} + P_{c0},
                     \\hfill 0.0^\\circ C \\leq T \\leq T_{max}
                    }

        Snow Density:
            .. math::
                \\rho_{s} = 50.0 + 1.7 * (T_{pp} + 15.0)^{ex}


                ex = \\Bigg \\lbrace{
                        ex_{min} + \\frac{T_{range} + T_{snow} - T_{max}}
                        {T_{range}} * ex_{r}, \\hfill ex < 1.75
                        \\atop
                        1.75, \\hfill, ex > 1.75
                        }

    """
    EX_MAX = 1.75
    EXR = 0.75
    EX_MIN = 1.0

    TZ = 0.0

    @staticmethod
    def run(tpp, _precipitation, t_max=0.0, t_min=-10.0, check_temps=True):
        """
        Args:
            tpp: A numpy array of temperature, use dew point temperature
                if available [degree C].
            _precipitation: A numpy array of precipitation in millimeters.
                Currently unused.
            t_max: Max temperature that snow density is modeled.
                   Default is 0.0 Degrees C.
            t_min: Minimum temperature that snow density is changing.
                   Default is -10.0 Degrees C.
            check_temps: A boolean value check to apply special temperature
                constraints, this is done using
                :func:`~smrf.envphys.Snow.check_temperature`. Default is True.

        Returns:
            Dictionary:
               - **pcs** (*numpy.array*) - Percent of the precipitation that is
                    snow in values 0.0-1.0.
               - **rho_s** (*numpy.array*) - Density of the fresh snow
                    in kg/m^3.
        """
        # again, this shouldn't be needed in this context
        if check_temps:
            tpp, tsnow = check_temperature(
                tpp, t_max=t_max, t_min=t_min
            )

        pcs = calc_percent_snow(tpp, t_max=t_max)

        # New snow density - no compaction
        t_range = t_max - t_min
        ex = PiecewiseSusong1999.EX_MIN + (
            ((t_range + (tsnow - t_max)) / t_range) * PiecewiseSusong1999.EXR
        )

        ex[ex > PiecewiseSusong1999.EX_MAX] = PiecewiseSusong1999.EX_MAX

        rho_ns = 50.0 + (1.7 * ((tpp - PiecewiseSusong1999.TZ) + 15.0)**ex)

        return {
            'pcs': pcs,
            'rho_s': rho_ns
        }



import os

from smrf import __version__
from smrf.framework.model_framework import SMRF
from smrf.tests.smrf_test_case import SMRFTestCase
from smrf.utils.utils import backup_input


class TestBackupInput(SMRFTestCase):

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.smrf = SMRF(cls.config_file)
        cls.smrf.loadTopo()
        cls.smrf.create_distribution()
        cls.smrf.loadData()

    def test_backup_version(self):

        backup_input(self.smrf.data, self.base_config)

        backup_config = os.path.join(
            self.base_config.cfg['output']['out_location'],
            'input_backup',
            'backup_config.ini')

        with open(backup_config) as f:
            lines = f.readlines()
            self.assertTrue(__version__ in lines[1])



import unittest

import netCDF4 as nc
import numpy as np

from smrf.data import Topo
from smrf.tests.smrf_test_case import SMRFTestCase


class TestLoadTopo(unittest.TestCase):
    @classmethod
    def setUp(self):
        topo_config = {
            'filename': SMRFTestCase.basin_dir.joinpath('topo', 'topo.nc'),
            'northern_hemisphere': True,
            'gradient_method': 'gradient_d8',
            'sky_view_factor_angles': 72
        }

        self.ds = nc.Dataset(topo_config['filename'])

        self.topo = Topo(topo_config)

    @classmethod
    def tearDown(self):
        self.ds.close()

    def test_center_calc_masked(self):
        '''
        Test the basin center calculation using the basin mask
        '''
        cx, cy = self.topo.get_center(self.ds, mask_name='mask')
        np.testing.assert_almost_equal(cx, 520033.7187500, 7)
        np.testing.assert_almost_equal(cy, 4768035.0, 7)

    def test_center_calc_domain(self):
        '''
        Test the basin center calculation for the entire basin domain
        '''
        cx, cy = self.topo.get_center(self.ds, mask_name=None)
        np.testing.assert_almost_equal(cx, 520050.0, 7)
        np.testing.assert_almost_equal(cy, 4768055.0, 7)

    def test_auto_calc_lat_lon(self):
        '''
        Test calculating the basin lat long correctly
        '''
        # Original RME
        # basin_lon:                     -116.7547
        # basin_lat:                     43.067
        np.testing.assert_almost_equal(
            self.topo.basin_lat, 43.06475372378507, 7)
        np.testing.assert_almost_equal(
            self.topo.basin_long, -116.75395420397061, 7)

    def test_projection_attributes(self):
        '''
        Confirm that this class has important projection attributes
        '''
        # Attribute directly used in load Grid as attributess from topo class
        important = ['basin_lat', 'basin_long', 'zone_number',
                     'northern_hemisphere']

        for at in important:
            self.assertTrue(hasattr(self.topo, at))



# -*- coding: utf-8 -*-
#
# SMRF documentation build configuration file, created by
# sphinx-quickstart on Sun Jul  3 08:08:09 2016.
#
# This file is execfile()d with the current directory set to its
# containing dir.
#
# Note that not all possible configuration values are present in this
# autogenerated file.
#
# All configuration values have a default; values that are commented out
# serve to show the default.

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.

import os
import sys
from datetime import datetime
# -- Have to do a mock install of some modules that RTD doesn't have --------
from unittest.mock import Mock

from inicheck.tools import config_documentation
from pkg_resources import get_distribution

import smrf
from smrf.utils.utils import get_config_doc_section_hdr

config_documentation('./user_guide/auto_config.rst',
                     modules='smrf',
                     section_link_dict=get_config_doc_section_hdr())

if os.environ.get('READTHEDOCS') == 'True':
    sys.path.insert(0, os.path.abspath('.'))
else:
    sys.path.insert(0, os.path.abspath('../'))


# class Mock(MagicMock):
#     @classmethod
#     def __getattr__(cls, name):
#             return Mock()
MOCK_MODULES = ['netCDF4', 'xarray', 'pandas']

sys.modules.update((mod_name, Mock()) for mod_name in MOCK_MODULES)

# -- General configuration ------------------------------------------------

# If your documentation needs a minimal Sphinx version, state it here.
#
# needs_sphinx = '1.0'

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    'sphinx.ext.doctest',
    'sphinx.ext.autodoc',
    'sphinx.ext.todo',
    'sphinx.ext.viewcode',
    'sphinx.ext.napoleon',
    'sphinx.ext.imgmath',
    'sphinxcontrib.bibtex',
    'sphinx.ext.intersphinx'
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# The suffix(es) of source filenames.
# You can specify multiple suffix as a list of string:
#
# source_suffix = ['.rst', '.md']
source_suffix = '.rst'

# The encoding of source files.
#
# source_encoding = 'utf-8-sig'

# The master toctree document.
master_doc = 'index'

# General information about the project.
project = 'SMRF'
copyright = f"2008-{datetime.now().year} USDA Agricultural Research Service"
author = 'Scott Havens'
# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.
#
# The short X.Y version.
version = get_distribution('smrf').version

# The full version, including alpha/beta/rc tags.
release = get_distribution('smrf').version

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#
# This is also used if you do content translation via gettext catalogs.
# Usually you set "language" from the command line for these cases.
language = None

# There are two options for replacing |today|: either, you set today to some
# non-false value, then it is used:
#
# today = ''
#
# Else, today_fmt is used as the format for a strftime call.
#
# today_fmt = '%B %d, %Y'

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This patterns also effect to html_static_path and html_extra_path
exclude_patterns = []

# The reST default role (used for this markup: `text`) to use for all
# documents.
#
# default_role = None

# If true, '()' will be appended to :func: etc. cross-reference text.
#
# add_function_parentheses = True

# If true, the current module name will be prepended to all description
# unit titles (such as .. function::).
#
# add_module_names = True

# If true, sectionauthor and moduleauthor directives will be shown in the
# output. They are ignored by default.
#
# show_authors = False

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = 'sphinx'

# A list of ignored prefixes for module index sorting.
# modindex_common_prefix = []

# If true, keep warnings as "system message" paragraphs in the built documents.
# keep_warnings = False

# If true, `todo` and `todoList` produce output, else they produce nothing.
todo_include_todos = False

# number the figures
numfig = True

# -- Google style ---------------------------------------------------------
# Napoleon settings
napoleon_google_docstring = True
napoleon_numpy_docstring = False
napoleon_include_init_with_doc = False
napoleon_include_private_with_doc = False
napoleon_include_special_with_doc = False
napoleon_use_admonition_for_examples = False
napoleon_use_admonition_for_notes = False
napoleon_use_admonition_for_references = False
napoleon_use_ivar = False
napoleon_use_param = True
napoleon_use_rtype = True
napoleon_use_keyword = True


# -- Options for HTML output ----------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = 'pydata_sphinx_theme'

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
#
# html_theme_options = {}

# Add any paths that contain custom themes here, relative to this directory.
# html_theme_path = []

# The name for this set of Sphinx documents.
# "<project> v<release> documentation" by default.
#
# html_title = u'SMRF v0.9.1'


# A shorter title for the navigation bar.  Default is the same as html_title.
#
# html_short_title = None

# The name of an image file (relative to this directory) to place at the top
# of the sidebar.
#
html_logo = '_static/ars_logo.png'

# The name of an image file (relative to this directory) to use as a favicon of
# the docs.  This file should be a Windows icon file (.ico) being 16x16 or 32x32
# pixels large.
#
# html_favicon = None

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static']

# Add any extra paths that contain custom files (such as robots.txt or
# .htaccess) here, relative to this directory. These files are copied
# directly to the root of the documentation.
#
# html_extra_path = []

# If not None, a 'Last updated on:' timestamp is inserted at every page
# bottom, using the given strftime format.
# The empty string is equivalent to '%b %d, %Y'.
#
# html_last_updated_fmt = None

# If true, SmartyPants will be used to convert quotes and dashes to
# typographically correct entities.
#
# html_use_smartypants = True

# Custom sidebar templates, maps document names to template names.
#
html_sidebars = {'**': ['globaltoc.html',
                        'relations.html', 'sourcelink.html', 'searchbox.html']}


# Output file base name for HTML help builder.
htmlhelp_basename = 'SMRFdoc'


# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title,
#  author, documentclass [howto, manual, or own class]).
latex_documents = [
    (master_doc, 'SMRF.tex', u'SMRF Documentation',
     u'Scott Havens', 'manual'),
]

# -- Options for manual page output ---------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [
    (master_doc, 'smrf', u'SMRF Documentation',
     [author], 1)
]

# If true, show URL addresses after external links.
#
# man_show_urls = False



import math

import numpy as np

from smrf.utils.io import isint

# define some constants
MAXV = 1.0              # vis albedo when gsize = 0
MAXIR = 0.85447         # IR albedo when gsize = 0
IRFAC = -0.02123        # IR decay factor
VFAC = 500.0            # visible decay factor
VZRG = 1.375e-3         # vis zenith increase range factor
IRZRG = 2.0e-3          # ir zenith increase range factor
IRZ0 = 0.1              # ir zenith increase range, gsize=0
BOIL = 373.15           # boiling temperature K
GRAVITY = 9.80665       # gravity (m/s^2)


def growth(t):
    """
    Calculate grain size growth
    From IPW albedo > growth
    """

    a = 4.0
    b = 3.
    c = 2.0
    d = 1.0

    factor = (a+(b*t)+(t*t))/(c+(d*t)+(t*t)) - 1.0

    return(1.0 - factor)


def albedo(telapsed, cosz, gsize, maxgsz, dirt=2):
    """
    Calculate the abedo, adapted from IPW function albedo

    Args:
        telapsed - time since last snow storm (decimal days)
        cosz - cosine local solar illumination angle matrix
        gsize - gsize is effective grain radius of snow after last storm (mu m)
        maxgsz - maxgsz is maximum grain radius expected from grain growth
                  (mu m)
        dirt - dirt is effective contamination for adjustment to visible
               albedo (usually between 1.5-3.0)

    Returns:
        tuple:
        Returns a tuple containing the visible and IR spectral albedo

        - **alb_v** (*numpy.array*) - albedo for visible specturm

        - **alb_ir** (*numpy.array*) -  albedo for ir spectrum

    Created April 17, 2015
    Modified July 23, 2015 - take image of cosz and calculate albedo for
        one time step
    Scott Havens

    """

#     telapsed = np.array(telapsed)

    # check inputs
    if gsize <= 0 or gsize > 500:
        raise Exception("unrealistic input: gsize=%i", gsize)

    if (maxgsz <= gsize or maxgsz > 2000):
        raise Exception("unrealistic input: maxgsz=%i", maxgsz)

    if 1 >= dirt >= 10:
        raise Exception("unrealistic input: dirt=%i", dirt)

    # set initial grain radii for vis and ir
    radius_ir = math.sqrt(gsize)
    range_ir = math.sqrt(maxgsz) - radius_ir
    radius_v = math.sqrt(dirt * gsize)
    range_v = math.sqrt(dirt * maxgsz) - radius_v

    # calc grain growth decay factor
    growth_factor = growth(telapsed + 1.0)

    # calc effective gsizes for vis & ir
    gv = radius_v + (range_v * growth_factor)
    gir = radius_ir + (range_ir * growth_factor)

    # calc albedos for cos(z)=1
    alb_v_1 = MAXV - (gv / VFAC)
    alb_ir_1 = MAXIR * np.exp(IRFAC * gir)

    # calculate effect of cos(z)<1

    # adjust diurnal increase range
    dzv = gv * VZRG
    dzir = (gir * IRZRG) + IRZ0

    # calculate albedo
    alb_v = alb_v_1
    alb_ir = alb_ir_1

    # correct if the sun is up
    ind = cosz > 0.0
    alb_v[ind] += dzv[ind] * (1.0 - cosz[ind])
    alb_ir[ind] += dzir[ind] * (1.0 - cosz[ind])

    return alb_v, alb_ir


def decay_alb_power(veg, veg_type, start_decay, end_decay,
                    t_curr, pwr, alb_v, alb_ir):
    """
    Find a decrease in albedo due to litter acccumulation. Decay is based on
    max decay, decay power, and start and end dates. No litter decay occurs
    before start_date. Fore times between start and end of decay,

    .. math::
      \\alpha = \\alpha - (dec_{max}^{\\frac{1.0}{pwr}} \\times
      \\frac{t-start}{end-start})^{pwr}

    Where :math:`\\alpha` is albedo, :math:`dec_{max}` is the maximum decay
    for albedo, :math:`pwr` is the decay power, :math:`t`, :math:`start`,
    and :math:`end` are the current, start, and end times for the litter decay.

    Args:
        start_decay: date to start albedo decay (datetime)
        end_decay: date at which to end albedo decay curve (datetime)
        t_curr: datetime object of current timestep
        pwr: power for power law decay
        alb_v: numpy array of albedo for visibile spectrum
        alb_ir: numpy array of albedo for IR spectrum

    Returns:
        tuple:
        Returns a tuple containing the corrected albedo arrays
        based on date, veg type
        - **alb_v** (*numpy.array*) - albedo for visible specturm

        - **alb_ir** (*numpy.array*) -  albedo for ir spectrum


    Created July 18, 2017
    Micah Sandusky

    """
    # Calculate hour past start of decay
    t_diff_hr = t_curr - start_decay
    t_diff_hr = t_diff_hr.days*24.0 + \
        t_diff_hr.seconds/3600.0  # only need hours here

    # Calculate total time of decay
    t_decay_hr = (end_decay - start_decay)
    t_decay_hr = t_decay_hr.days*24.0 + \
        t_decay_hr.seconds/3600.0  # only need hours here

    # correct for veg
    alb_dec = np.zeros_like(alb_v)

    # Don't decay if before start
    if t_diff_hr <= 0.0:
        alb_dec = alb_dec * 0.0

    # Use max decay if after start
    elif t_diff_hr > t_decay_hr:
        # Use default
        alb_dec = alb_dec + veg['default']
        # Decay based on veg type
        for k, v in veg.items():
            if isint(k):
                alb_dec[veg_type == int(k)] = v

    # Power function decay if during decay period
    else:
        # Use defaults
        max_dec = veg['default']
        tao = (t_decay_hr) / (max_dec**(1.0/pwr))

        # Add default decay to array of zeros
        alb_dec = alb_dec + ((t_diff_hr) / tao)**pwr

        # Decay based on veg type
        for k, v in veg.items():
            max_dec = v
            tao = (t_decay_hr) / (max_dec**(1.0/pwr))

            # Set albedo decay at correct veg types
            if isint(k):
                alb_dec[veg_type == int(k)] = ((t_diff_hr) / tao)**pwr

    alb_v_d = alb_v - alb_dec
    alb_ir_d = alb_ir - alb_dec

    return alb_v_d, alb_ir_d


def decay_alb_hardy(litter, veg_type, storm_day, alb_v, alb_ir):
    """
    Find a decrease in albedo due to litter acccumulation
    using method from :cite:`Hardy:2000` with storm_day as input.

    .. math::
        lc = 1.0 - (1.0 - lr)^{day}

    Where :math:`lc` is the fractional litter coverage and :math:`lr` is
    the daily litter rate of the forest. The new albedo is a weighted
    average of the calculated albedo for the clean snow and the albedo
    of the litter.

    Note: uses input of l_rate (litter rate) from config
    which is based on veg type. This is decimal percent litter
    coverage per day

    Args:
        litter: A dictionary of values for default,albedo,41,42,43 veg types
        veg_type: An image of the basin's NLCD veg type
        storm_day: numpy array of decimal day since last storm
        alb_v: numpy array of albedo for visibile spectrum
        alb_ir: numpy array of albedo for IR spectrum
        alb_litter: albedo of pure litter

    Returns:
        tuple:
        Returns a tuple containing the corrected albedo arrays
        based on date, veg type
        - **alb_v** (*numpy.array*) - albedo for visible specturm

        - **alb_ir** (*numpy.array*) -  albedo for ir spectrum

    Created July 19, 2017
    Micah Sandusky

    """
    # array for decimal percent snow coverage
    sc = np.zeros_like(alb_v)
    # calculate snow coverage default veg type
    l_rate = litter['default']
    alb_litter = litter['albedo']

    sc = sc + (1.0-l_rate)**(storm_day)

    # calculate snow coverage based on veg type
    for k, v in litter.items():

        l_rate = litter[k]
        if isint(k):
            sc[veg_type == int(k)] = (
                1.0 - l_rate)**(storm_day[veg_type == int(k)])

    # calculate litter coverage
    lc = np.ones_like(alb_v) - sc

    # weighted average to find decayed albedo
    alb_v_d = alb_v*sc + alb_litter*lc
    alb_ir_d = alb_ir*sc + alb_litter*lc

    return alb_v_d, alb_ir_d



import os

import netCDF4 as nc
import numpy as np

from smrf import __version__
from smrf.framework.model_framework import SMRF
from smrf.output.output_netcdf import OutputNetcdf
from smrf.tests.smrf_test_case import SMRFTestCase


class TestOutputNetCDF(SMRFTestCase):

    def setUp(self):
        super().setUpClass()
        self.smrf = SMRF(self.config_file)

        self.variable_dict = {
            'air_temp': {
                'variable': 'air_temp',
                'module': 'air_temp',
                'out_location': os.path.join(
                    self.smrf.config['output']['out_location'],
                    'air_temp'
                ),
                'info': {
                    'units': 'degree_Celsius',
                    'standard_name': 'air_temperature',
                    'long_name': 'Air temperature'
                }
            }
        }

    def test_netcdf_smrf_version(self):

        self.smrf.loadTopo()

        OutputNetcdf(
            self.variable_dict,
            self.smrf.topo,
            self.smrf.config['time'],
            self.smrf.config['output'])

        n = nc.Dataset(self.variable_dict['air_temp']['file_name'])
        version = n.getncattr('SMRF_version')
        self.assertTrue(n.variables['air_temp'].dtype == np.float32)
        n.close()

        self.assertEqual(__version__, version)

    def test_netcdf_precision(self):
        self.smrf.config['output']['netcdf_output_precision'] = 'double'

        self.smrf.loadTopo()

        OutputNetcdf(
            self.variable_dict,
            self.smrf.topo,
            self.smrf.config['time'],
            self.smrf.config['output'])

        n = nc.Dataset(self.variable_dict['air_temp']['file_name'])
        self.assertTrue(n.variables['air_temp'].dtype == np.float64)
        n.close()




import logging

import numpy as np
from netCDF4 import Dataset
from topocalc import gradient
from topocalc.viewf import viewf
from utm import to_latlon


class Topo():
    """
    Class for topo images and processing those images. Images are:
    - DEM
    - Mask
    - veg type
    - veg height
    - veg k
    - veg tau

    Inputs to topo are the topo section of the config file

    """

    IMAGES = ['dem', 'mask', 'veg_type', 'veg_height', 'veg_k', 'veg_tau']

    def __init__(self, topoConfig):
        self.topoConfig = topoConfig

        self._logger = logging.getLogger(__name__)
        self._logger.info('Reading [TOPO] and making stoporad input')

        self.readNetCDF()

        # calculate the gradient and the sky view factor
        self.gradient()
        self.viewf()

    def readNetCDF(self):
        """
        Read in the images from the config file where the file
        listed is in netcdf format
        """

        # read in the images
        f = Dataset(self.topoConfig['filename'], 'r')

        # netCDF>1.4.0 returns as masked arrays even if no missing values
        # are present. This will ensure that if the array has no missing
        # values, a normal numpy array is returned
        f.set_always_mask(False)

        if 'projection' not in f.variables.keys():
            raise IOError("Topo input files must have projection information")

        self.readImages(f)

        # get some general information about the model domain from the dem
        self.nx = f.dimensions['x'].size
        self.ny = f.dimensions['y'].size

        # create the x,y vectors
        self.x = f.variables['x'][:]
        self.y = f.variables['y'][:]
        [self.X, self.Y] = np.meshgrid(self.x, self.y)

        # There is not a great NetCDF convention on direction for the y-axis.
        # So there is the possibility that the dy will be positive or negative.
        # For the gradient calculations this needs to be absolute spacing.
        self.dx = np.mean(np.diff(self.x))
        self.dy = np.abs(np.mean(np.diff(self.y)))

        # Calculate the center of the basin
        self.cx, self.cy = self.get_center(f, mask_name='mask')

        # Is the modeling domain in the northern hemisphere
        self.northern_hemisphere = self.topoConfig['northern_hemisphere']

        # Assign the UTM zone
        self.zone_number = int(f.variables['projection'].utm_zone_number)

        # Calculate the lat long
        self.basin_lat, self.basin_long = to_latlon(
            self.cx,
            self.cy,
            self.zone_number,
            northern=self.northern_hemisphere)

        self._logger.info('Domain center in UTM Zone {:d} = {:0.1f}m, {:0.1f}m'
                          ''.format(self.zone_number, self.cx, self.cy))
        self._logger.info('Domain center as Latitude/Longitude = {:0.5f}, '
                          '{:0.5f}'.format(self.basin_lat, self.basin_long))

        f.close()

    def readImages(self, f):
        """Read images from the netcdf and set as attributes in the Topo class

        Args:
            f: netcdf dataset object
        """

        # netCDF files are stored typically as 32-bit float, so convert
        # to double or int
        for v_smrf in self.IMAGES:

            if v_smrf in f.variables.keys():
                if v_smrf == 'veg_type':
                    result = f.variables[v_smrf][:].astype(int)
                else:
                    result = f.variables[v_smrf][:].astype(np.float64)

            setattr(self, v_smrf, result)

    def get_center(self, ds, mask_name=None):
        '''
        Function returns the basin center in the native coordinates of the
        a netcdf object.

        The incoming data set must contain at least and x, y and optionally
        whatever mask name the user would like to use for calculating .
        If no mask name is provided then the entire domain is used.

        Args:
            ds: netCDF4.Dataset object containing at least x,y, optionally
                    a mask variable name
            mask_name: variable name in the dataset that is a mask where 1 is
                    in the mask
        Returns:
            tuple: x,y of the data center in the datas native coordinates
        '''
        x = ds.variables['x'][:]
        y = ds.variables['y'][:]

        # Calculate the center of the basin
        if mask_name is not None:
            mask_id = np.argwhere(ds.variables[mask_name][:] == 1)

            # Tuple is required for an upcoming deprecation in numpy
            idx = tuple([mask_id[:, 1]])
            idy = tuple([mask_id[:, 0]])

            x = x[idx]
            y = y[idy]

        return x.mean(), y.mean()

    def gradient(self):
        """
        Calculate the gradient and aspect
        """

        func = self.topoConfig['gradient_method']

        # calculate the gradient and aspect
        g, a = getattr(gradient, func)(
            self.dem, self.dx, self.dy, aspect_rad=True)
        self.slope_radians = g

        # following IPW convention for slope as sin(Slope)
        self.sin_slope = np.sin(g)
        self.aspect = a

    def viewf(self):
        """Calculate the sky view factor
        """

        svf, tcf = viewf(
            self.dem,
            self.dx,
            nangles=self.topoConfig['sky_view_factor_angles'],
            sin_slope=self.sin_slope,
            aspect=self.aspect)

        self.sky_view_factor = svf
        self.terrain_config_factor = tcf



import numpy as np


def solar_veg_beam(data, height, cosz, k):
    """
    Apply the vegetation correction to the beam irradiance
    using the equation from Links and Marks 1999

    S_b,f = S_b,o * exp[ -k h sec(theta) ] or
    S_b,f = S_b,o * exp[ -k h / cosz ]

    20150610 Scott Havens
    """

    # ensure that the sun is visible
    cosz[cosz <= 0] = 0.01

    return data * np.exp(-k * height / cosz)


def solar_veg_diffuse(data, tau):
    """
    Apply the vegetation correction to the diffuse irradiance
    using the equation from Links and Marks 1999

    S_d,f = tau * S_d,o

    20150610 Scott Havens
    """

    return tau * data



'''
2016-03-07 Scott Havens

Distributed forcing data over a grid using interpolation
'''

import numpy as np
import pandas as pd
from scipy.interpolate import griddata
from scipy.interpolate.interpnd import _ndim_coords_from_arrays
from scipy.spatial import qhull as qhull

from smrf.utils.utils import grid_interpolate_deconstructed


class GRID:
    '''
    Inverse distance weighting class
    - Standard IDW
    - Detrended IDW
    '''

    def __init__(self, config, mx, my, GridX, GridY,
                 mz=None, GridZ=None, mask=None, metadata=None):
        """
        Args:
            config: configuration for grid interpolation
            mx: x locations for the points
            my: y locations for the points
            mz: z locations for the points
            GridX: x locations in grid to interpolate over
            GridY: y locations in grid to interpolate over
            GridZ: z locations in grid to interpolate over
            mask: mask for those points to include in the detrending
                will be ignored if config['mask'] is false
        """

        self.config = config

        # measurement point locations
        self.mx = mx
        self.my = my
        self.mz = mz
        self.npoints = len(mx)

        # grid information
        self.GridX = GridX
        self.GridY = GridY
        self.GridZ = GridZ

        self.metadata = metadata

        # local elevation gradient, precalculte the distance dataframe
        if config['grid_local']:
            k = config['grid_local_n']
            dist_df = pd.DataFrame(index=metadata.index, columns=range(k))
            dist_df.index.name = 'cell_id'
            for i, row in metadata.iterrows():

                d = np.sqrt((metadata.latitude - row.latitude) **
                            2 + (metadata.longitude - row.longitude)**2)
                dist_df.loc[row.name] = d.sort_values()[:k].index

            self.dist_df = dist_df

            # stack and reset index
            df = dist_df.stack().reset_index()
            df = df.rename(columns={0: 'cell_local'})
            df.drop('level_1', axis=1, inplace=True)

            # get the elevations
            df['elevation'] = metadata.loc[df.cell_local, 'elevation'].values

            # now we have cell_id, cell_local and elevation for the whole grid
            self.full_df = df

            self.tri = None

        # mask
        self.mask = np.zeros_like(self.mx, dtype=bool)
        if config['grid_mask']:

            assert(mask.shape == GridX.shape)
            mask = mask.astype(bool)

            x = GridX[0, :]
            y = GridY[:, 0]
            for i, v in enumerate(mx):
                xi = np.argmin(np.abs(x - mx[i]))
                yi = np.argmin(np.abs(y - my[i]))

                self.mask[i] = mask[yi, xi]
        else:
            self.mask = np.ones_like(self.mx, dtype=bool)

    def detrendedInterpolation(self, data, flag=0, grid_method='linear'):
        """
        Interpolate using a detrended approach

        Args:
            data: data to interpolate
            grid_method: scipy.interpolate.griddata interpolation method
        """

        # get the trend, ensure it's positive

        if self.config['grid_local']:
            rtrend = self.detrendedInterpolationLocal(data, flag, grid_method)

        else:
            rtrend = self.detrendedInterpolationMask(data, flag, grid_method)

        return rtrend

    def detrendedInterpolationLocal(self, data, flag=0, grid_method='linear'):
        """
        Interpolate using a detrended approach

        Args:
            data: data to interpolate
            grid_method: scipy.interpolate.griddata interpolation method
        """

        # take the new full_df and fill a data column
        df = self.full_df.copy()
        df['data'] = data[df['cell_local']].values
        df = df.set_index('cell_id')
        df['fit'] = df.groupby('cell_id').apply(
            lambda x: np.polyfit(x.elevation, x.data, 1))

        # drop all the duplicates
        df.reset_index(inplace=True)
        df.drop_duplicates(subset=['cell_id'], keep='first', inplace=True)
        df.set_index('cell_id', inplace=True)
        df[['slope', 'intercept']] = df.fit.apply(pd.Series)
        # df = df.drop(columns='fit').reset_index()

        # apply trend constraints
        if flag == 1:
            df.loc[df['slope'] < 0, ['slope', 'intercept']] = 0
        elif flag == -1:
            df.loc[df['slope'] > 0, ['slope', 'intercept']] = 0

        # get triangulation
        if self.tri is None:
            xy = _ndim_coords_from_arrays(
                (self.metadata.utm_x, self.metadata.utm_y))
            self.tri = qhull.Delaunay(xy)

        # interpolate the slope/intercept
        grid_slope = grid_interpolate_deconstructed(
            self.tri,
            df.slope.values[:],
            (self.GridX, self.GridY),
            method=grid_method)

        grid_intercept = grid_interpolate_deconstructed(
            self.tri,
            df.intercept.values[:],
            (self.GridX, self.GridY),
            method=grid_method)

        # remove the elevation trend from the HRRR precip
        el_trend = df.elevation * df.slope + df.intercept
        dtrend = df.data - el_trend

        # interpolate the residuals over the DEM
        idtrend = grid_interpolate_deconstructed(
            self.tri,
            dtrend,
            (self.GridX, self.GridY),
            method=grid_method)

        # reinterpolate
        rtrend = idtrend + grid_slope * self.GridZ + grid_intercept

        return rtrend

    def detrendedInterpolationMask(self, data, flag=0, grid_method='linear'):
        """
        Interpolate using a detrended approach

        Args:
            data: data to interpolate
            grid_method: scipy.interpolate.griddata interpolation method
        """

        # get the trend, ensure it's positive
        pv = np.polyfit(self.mz[self.mask].astype(float), data[self.mask], 1)

        # apply trend constraints
        if flag == 1 and pv[0] < 0:
            pv = np.array([0, 0])
        elif (flag == -1 and pv[0] > 0):
            pv = np.array([0, 0])

        self.pv = pv

        # detrend the data
        el_trend = self.mz * pv[0] + pv[1]
        dtrend = data - el_trend

        # interpolate over the DEM grid
        idtrend = griddata((self.mx, self.my),
                           dtrend,
                           (self.GridX, self.GridY),
                           method=grid_method)

        # retrend the data
        rtrend = idtrend + pv[0]*self.GridZ + pv[1]

        return rtrend

    def calculateInterpolation(self, data, grid_method='linear'):
        """
        Interpolate over the grid

        Args:
            data: data to interpolate
            mx: x locations for the points
            my: y locations for the points
            X: x locations in grid to interpolate over
            Y: y locations in grid to interpolate over
        """

        g = griddata((self.mx, self.my),
                     data,
                     (self.GridX, self.GridY),
                     method=grid_method)

        return g



import os

import numpy as np
import pandas as pd
from topocalc.shade import shade

from smrf.data import Topo
from smrf.envphys.albedo import albedo
from smrf.envphys.solar import irradiance, toporad
from smrf.envphys.sunang import sunang
from smrf.tests.smrf_test_case_lakes import SMRFTestCaseLakes


class TestToporad(SMRFTestCaseLakes):

    @classmethod
    def setUpClass(cls):
        super().setUpClass()

        date_time = pd.to_datetime('2/15/1990 20:30')
        cls.date_time = date_time.tz_localize('UTC')

        cls.tau_elevation = 100.0
        cls.tau = 0.2
        cls.omega = 0.85
        cls.scattering_factor = 0.3
        cls.surface_albedo = 0.5
        cls.solar_irradiance = irradiance.direct_solar_irradiance(
            cls.date_time, w=[0.28, 2.8])

        topo_config = {
            'filename': os.path.join(cls.basin_dir, 'topo/topo.nc'),
            'northern_hemisphere': True,
            'gradient_method': 'gradient_d8',
            'sky_view_factor_angles': 72
        }
        cls.topo = Topo(topo_config)
        cls.dem = cls.topo.dem

        # inputs for toporad and stoporad
        cls.cosz, cls.azimuth, rad_vec = sunang(
            cls.date_time,
            cls.topo.basin_lat,
            cls.topo.basin_long)

        cls.elevrad = toporad.Elevrad(
            cls.dem,
            cls.solar_irradiance,
            cls.cosz)

        cls.illum_ang = shade(
            cls.topo.sin_slope,
            cls.topo.aspect,
            cls.azimuth,
            cls.cosz)

    def test_elevrad(self):

        rad = toporad.Elevrad(
            self.dem,
            self.solar_irradiance,
            cosz=0.45)

        self.assertAlmostEqual(965.8828575567114, np.mean(rad.beam), places=4)
        self.assertAlmostEqual(943.7263794826135, np.min(rad.beam), places=4)
        self.assertAlmostEqual(989.6429308508227, np.max(rad.beam), places=4)

        self.assertAlmostEqual(
            93.84315440085071, np.mean(rad.diffuse), places=4)
        self.assertAlmostEqual(
            87.84988698685561, np.min(rad.diffuse), places=4)
        self.assertAlmostEqual(
            99.39603484739534, np.max(rad.diffuse), places=4)

    def test_elevrad_options(self):

        rad = toporad.Elevrad(
            self.dem,
            self.solar_irradiance,
            cosz=0.5,
            tau_elevation=80,
            tau=0.3,
            omega=0.55,
            scattering_factor=0.35,
            surface_albedo=0.3)

        self.assertAlmostEqual(866.5210770956645, np.mean(rad.beam), places=4)
        self.assertAlmostEqual(839.8354626316999, np.min(rad.beam), places=4)
        self.assertAlmostEqual(895.3273572653653, np.max(rad.beam), places=4)

        self.assertAlmostEqual(
            76.05068555235343, np.mean(rad.diffuse), places=4)
        self.assertAlmostEqual(
            71.93616622814766, np.min(rad.diffuse), places=4)
        self.assertAlmostEqual(
            79.77791625995671, np.max(rad.diffuse), places=4)

    def test_toporad(self):

        trad_beam, trad_diffuse = toporad.toporad(
            self.elevrad.beam,
            self.elevrad.diffuse,
            self.illum_ang,
            self.topo.sky_view_factor,
            self.topo.terrain_config_factor,
            self.cosz,
            surface_albedo=0.5)

        self.assertAlmostEqual(706.0067021452699, np.mean(trad_beam), places=4)
        self.assertAlmostEqual(85.17264068026927, np.min(trad_beam), places=4)
        self.assertAlmostEqual(1164.002943767303, np.max(trad_beam), places=4)

        self.assertAlmostEqual(
            112.40390734361924, np.mean(trad_diffuse), places=4)
        self.assertAlmostEqual(
            84.87509154016982, np.min(trad_diffuse), places=4)
        self.assertAlmostEqual(
            148.2553072510663, np.max(trad_diffuse), places=4)

    def test_stoporad_visible(self):

        alb_vis, alb_ir = albedo(
            20 * np.ones_like(self.dem), self.illum_ang, 500, 2000)

        srad_beam, srad_diffuse = toporad.stoporad(
            self.date_time,
            self.topo,
            self.cosz,
            self.azimuth,
            self.illum_ang,
            alb_vis,
            wavelength_range=[0.28, 0.7],
            tau_elevation=self.tau_elevation,
            tau=self.tau,
            omega=self.omega,
            scattering_factor=self.scattering_factor)

        self.assertAlmostEqual(
            350.1547099434225, np.mean(srad_beam), places=4)
        self.assertAlmostEqual(49.930680972072594, np.min(srad_beam), places=4)
        self.assertAlmostEqual(570.8947805048191, np.max(srad_beam), places=4)

        self.assertAlmostEqual(
            65.79420200041265, np.mean(srad_diffuse), places=4)
        self.assertAlmostEqual(
            49.930680972072594, np.min(srad_diffuse), places=4)
        self.assertAlmostEqual(
            102.02449905038313, np.max(srad_diffuse), places=4)

    def test_stoporad_ir(self):

        alb_vis, alb_ir = albedo(
            20 * np.ones_like(self.dem), self.illum_ang, 500, 2000)

        srad_beam, srad_diffuse = toporad.stoporad(
            self.date_time,
            self.topo,
            self.cosz,
            self.azimuth,
            self.illum_ang,
            alb_ir,
            wavelength_range=[0.7, 2.8],
            tau_elevation=self.tau_elevation,
            tau=self.tau,
            omega=self.omega,
            scattering_factor=self.scattering_factor)

        self.assertAlmostEqual(
            365.23243248652074, np.mean(srad_beam), places=4)
        self.assertAlmostEqual(41.8954631805951, np.min(srad_beam), places=4)
        self.assertAlmostEqual(602.1472248320969, np.max(srad_beam), places=4)

        self.assertAlmostEqual(
            56.30765350042961, np.mean(srad_diffuse), places=4)
        self.assertAlmostEqual(
            41.8954631805951, np.min(srad_diffuse), places=4)
        self.assertAlmostEqual(
            71.27409494983297, np.max(srad_diffuse), places=4)



import logging
from math import asin, atan, cos, fmod, sin, sqrt, tan

import numpy as np
import pytz

JULIAN_CENTURY = 36525		# days in Julian century
DEGS_IN_CIRCLE = 3.6e2        # degrees in circle
TOLERANCE = 1.1920928955e-07
M_PI = 3.14159265358979323846
M_PI_2 = 2 * M_PI
MAX_DECLINATION = 23.6


def sunang(date_time, latitude, longitude, truncate=True):
    """
    Calculate the sun angle (the azimuth and zenith angles of
    the sun's position) for a given geodetic location for a single
    date time and coordinates. The function can take either latitude
    longitude position as a single point or numpy array.

    Args:
        date_time: python datetime object
        latitude: value or np.ndarray (in degrees)
        longitude: value or np.ndarray (in degrees)
        truncate: True will replicate the IPW output precision, not
                applied if position is an array

    Returns
        cosz - cosine of the zenith angle, same shape as input position
        azimuth - solar azimuth, same shape as input position
        rad_vec - Earth-Sun radius vector

    """

    # Calculate the ephemeris parameters
    declination, omega, rad_vec = ephemeris(date_time)

    # calculate the sun angle
    rad_latitude = latitude * M_PI / 180
    rad_longitude = longitude * M_PI / 180
    mu, azimuth = sunpath(rad_latitude, rad_longitude, declination, omega)

    azimuth = azimuth * 180 / M_PI

    if truncate and not isinstance(mu, np.ndarray):
        mu = float('{0:.6f}'.format(mu))
        azimuth = float('{0:.5g}'.format(azimuth))
        rad_vec = float('{0:.5g}'.format(rad_vec))

    return mu, azimuth, rad_vec


def sunang_thread(queue, date, lat, lon):
    """
    See sunang for input descriptions

    Args:
        queue: queue with cosz, azimuth
        date: loop through dates to accesss queue, must be same as
                rest of queues

    """

    if 'cosz' not in queue.keys():
        raise ValueError('queue must have cosz key')
    if 'azimuth' not in queue.keys():
        raise ValueError('queue must have cosz key')

    log = logging.getLogger(__name__)

    for t in date:

        log.debug('%s Calculating sun angle' % t)

        cosz, azimuth, rad_vec = sunang(t.astimezone(pytz.utc), lat, lon)

        queue['cosz'].put([t, cosz])
        queue['azimuth'].put([t, azimuth])


def sunpath(latitude, longitude, declination, omega):
    """
    Sun angle from solar declination and longtitude

    Args:
        latitude: in radians
        longitude: in radians
        declination: solar declination (radians)
        omega: solar longitude (radians)

    Returns
        cosz: cosine of solar zenith
        azimuth: solar azimuth in radians
    """

    if isinstance(latitude, np.ndarray):
        if np.any(np.abs(latitude) > M_PI_2):
            raise ValueError("latitude array not between -90 and +90 degrees")

        if np.any(np.abs(longitude) > M_PI):
            raise ValueError(
                "longitude array not between -180 and +180 degrees")

    else:
        if np.abs(latitude) > M_PI_2:
            raise ValueError(
                "latitude ({} deg) not between -90 and +90 degrees".format(
                    latitude*180/M_PI))

        if np.abs(longitude) > M_PI:
            raise ValueError(
                "longitude ({} deg) not between -180 and +180 degrees".format(
                    longitude*180/M_PI))

    if np.abs(declination) > M_PI*MAX_DECLINATION/180:
        raise ValueError(
            "declination ({} deg) > maximum declination ({} deg)".format(
                declination*180/M_PI, MAX_DECLINATION))

    if np.abs(omega) > M_PI:
        raise ValueError(
            """longitude of sun ({} deg) not between -180 and """
            """+180 degrees""".format(omega*180/M_PI))

    cosz, az = rotate(np.sin(declination), omega, np.sin(latitude), longitude)

    return cosz, az


def dsign(a, b):
    """
    modified from /usr/src/lib/libF77/d_sign.c
    """

    x = a if a >= 0 else -a
    y = x if b >= 0 else -x
    return y

# original ibm 360/44 fortran ivf - vislab - wilson - 29jul79
# translated, modified and reduced by dozier - ucsb - 11/81


def ephemeris(dt):
    """
    Calculates radius vector, declination, and apparent longitude
    of sun, as function of the given date and time.

    The routine is adapted from:

    W. H. Wilson, Solar ephemeris algorithm, Reference 80-13, 70
        pp., Scripps Institution of Oceanography, University of
        California, San Diego, La Jolla, CA, 1980.

    Args:
        dt: date time python object

    Returns:
        declin: solar declination angle, in radians
        omega: sun longitude, in radians
        r: Earth-Sun radius vector

    """

    one = 1.0
    degrd = atan(one) / 45.0

    # Convert time to seconds since midnight
    gmts = 3600.0 * dt.hour + 60.0 * dt.minute + dt.second

    # p51 = gmts/3600/24*360

    # The int() is required for compatability with IPW as the divide by
    # 100 keeps the value as an integer where python converts to a float...
    p51 = gmts / 10.0 / 24.0
    p22 = int(((dt.year - 1900) * JULIAN_CENTURY - 25) / 100) + \
        yearday(dt.year, dt.month, dt.day) - 0.5
    p23 = (p51 / DEGS_IN_CIRCLE + p22) / JULIAN_CENTURY
    p22 = p23 * JULIAN_CENTURY

    # mean longitude - p24
    p11 = 279.69668
    p12 = 0.9856473354
    p13 = 3.03e-4
    p24 = p11 + fmod(p12 * p22, DEGS_IN_CIRCLE) + p13 * p23 * p23
    p24 = fmod(p24, DEGS_IN_CIRCLE)

    # mean anomaly - p25
    p11 = 358.47583
    p12 = 0.985600267
    p13 = -1.5e-4
    p14 = -3.e-6
    p25 = p11 + fmod(p12 * p22, DEGS_IN_CIRCLE) + p23 * p23 * (p13 + p14 * p23)
    p25 = fmod(p25, DEGS_IN_CIRCLE)

    # eccentricity - p26
    p11 = 0.01675104
    p12 = -4.18e-5
    p13 = -1.26e-7
    p26 = p11 + p23 * (p12 + p13 * p23)
    p11 = p25 * degrd
    p12 = p11

    while True:
        p13 = p12
        p12 = p11 + p26 * sin(p13)
        if abs((p12 - p13) / p12) < TOLERANCE:
            break

    p13 = p12 / degrd

    # true anomaly - p27`
    p27 = 2.0 * atan(sqrt((1.0 + p26) / (1.0 - p26))
                     * tan(p13 / 2.0 * degrd)) / degrd
    if dsign(1.0, p27) != dsign(1.0, sin(p13 * degrd)):
        p27 += 1.8e2
    if p27 < 0.0:
        p27 += DEGS_IN_CIRCLE

    # radius vector - r
    r = 1.0 - p26 * cos(p13 * degrd)

    # aberration - p29
    p29 = -20.47 / r / 3600

    # mean obliquity - p43
    p11 = 23.452294
    p12 = -0.0130125
    p13 = -1.64e-6
    p14 = 5.03e-7
    p43 = p11 + p23 * (p12 + p23 * (p13 + p14 * p23))

    #  mean ascension - p45
    p11 = 279.6909832
    p12 = 0.98564734
    p13 = 3.8707
    p13 = 3.8708e-4
    p45 = p11 + fmod(p12 * p22, DEGS_IN_CIRCLE) + p13 * p23 * p23
    p45 = fmod(p45, DEGS_IN_CIRCLE)

    # nutation and longitude pert
    p11 = 296.104608
    p12 = 1325 * DEGS_IN_CIRCLE
    p13 = 198.8491083
    p14 = 0.00919167
    p15 = 1.4388e-5
    p28 = p11 + fmod(p12 * p23, DEGS_IN_CIRCLE) + p23 * \
        (p13 + p23 * (p14 + p15 * p23))
    p28 = fmod(p28, DEGS_IN_CIRCLE)

    # mean elongation of moon - p30
    p11 = 350.737486
    p12 = 1236 * DEGS_IN_CIRCLE
    p13 = 307.1142167
    p14 = 1.436e-3
    p30 = p11 + fmod(p12 * p23, DEGS_IN_CIRCLE) + p23 * (p13 + p14 * p23)
    p30 = fmod(p30, DEGS_IN_CIRCLE)

    # moon long of ascending node - p31
    p11 = 259.183275
    p12 = -5 * DEGS_IN_CIRCLE
    p13 = -134.142008
    p14 = 2.0778e-3
    p31 = p11 + fmod(p12 * p23, DEGS_IN_CIRCLE) + p23 * (p13 + p14 * p23)
    p31 = fmod(p31, DEGS_IN_CIRCLE)

    # mean long of moon - p31
    p11 = 270.434164
    p12 = 1336 * DEGS_IN_CIRCLE
    p13 = 307.8831417
    p14 = -1.1333e-3
    p32 = p11 + fmod(p12 * p23, DEGS_IN_CIRCLE) + p23 * (p13 + p14 * p23)
    p32 = fmod(p32, DEGS_IN_CIRCLE)

    # moon perturbation of sun long - p33
    p33 = 6.454 * sin(p30 * degrd) + 0.013 * sin(3 * p30 * degrd) + \
        0.177 * sin((p30 + p28) * degrd) - \
        0.424 * sin((p30 - p28) * degrd)
    p33 = p33 / 3600

    # nutation of long - p34
    p34 = -(17.234 - 0.017 * p23) * sin(p31 * degrd) + \
        0.209 * sin(2 * p31 * degrd) - \
        0.204 * sin(2 * p32 * degrd)
    p34 = p34 - 1.257 * sin(2 * p24 * degrd) + 0.127 * sin(p28 * degrd)
    p34 = p34 / 3600

    # nutation in obliquity - p34
    p35 = 9.214 * cos(p31 * degrd) + 0.546 * cos(2 * p24 * degrd) - \
        .09 * cos(2 * p31 * degrd) + 0.088 * cos(2 * p32 * degrd)
    p35 = p35 / 3600

    # inequalities of long period - p36
    p36 = 0.266 * sin((31.8 + 119 * p23) * degrd) + \
        ((1.882 - 0.016 * p23) * degrd) * \
        sin((57.24 + 150.27 * p23) * degrd)
    p36 = p36 + 0.202 * sin((315.6 + 893.3 * p23) * degrd) + \
        1.089 * p23 * p23 + 6.4 * sin((231.19 + 20.2 * p23) * degrd)
    p36 = p36 / 3600

    # apparent longitude - p41
    p41 = p27 - p25 + p24 + p29 + p33 + p36 + p34

    p43 += p35

    # apparent right ascension - p44
    p44 = atan(tan(p41 * degrd) * cos(p43 * degrd)) / degrd
    if dsign(one, p44) != dsign(one, sin(p41 * degrd)):
        p44 += 1.8e2
    if p44 < 0:
        p44 += DEGS_IN_CIRCLE

    # equation of time - p46
    p46 = p45 - p44
    if p46 > 1.8e2:
        p46 -= DEGS_IN_CIRCLE

    # declination - p47
    p47 = asin(sin(p41 * degrd) * sin(p43 * degrd))
    declin = p47

    # hour angle in degrees - p48
    p48 = p51 + p46 - 1.8e2
    if p48 > 180:
        p48 -= DEGS_IN_CIRCLE
    elif p48 < -180:
        p48 += DEGS_IN_CIRCLE
    omega = -p48 * degrd

    return declin, omega, r


def rotate(mu, azm, mu_r, lam_r):
    """
    Calculates new spherical coordinates if system rotated about
    origin.  Coordinates are right-hand system.  All angles are in
    radians.

    Args:
        mu: cosine of angle theta from z-axis in old coordinate
            system, sin(declination)
        azm: azimuth (+ccw from x-axis) in old coordinate system,
            hour angle of sun (long. where sun is vertical)
        mu_r: cosine of angle theta_r of rotation of z-axis, sin(latitude)
        lam_r: azimuth (+ccw) of rotation of x-axis, longitude

    Returns:
        muPrime: cosine of the solar zenith
        aPrime: solar azimuth in radians
    """

    # Check input values: mu = cos(theta) mu_r = cos(theta-sub-r)
    if mu < -1.0 or mu > 1.0:
        raise ValueError(
            "rotate: mu = cos(theta) = {} is not between -1 and +1".format(mu))

    if isinstance(mu_r, np.ndarray):
        if np.any(mu_r < -1.0) or np.any(mu_r > 1.0):
            raise ValueError("rotate, mu_r array is not between -1 and +1")
    else:
        if mu_r < -1.0 or mu_r > 1.0:
            raise ValueError(
                "rotate: mu_r ({}) is not between -1 and +1".format(mu_r))

    if np.abs(azm) > (2 * np.pi):
        raise ValueError(
            "rotate: azimuth ({} deg) is not between -360 and 360".format(
                180 * (azm * 360) / np.pi))

    if isinstance(lam_r, np.ndarray):
        if np.any(np.abs(lam_r) > (2 * np.pi)):
            raise ValueError("rotate, lam_r array is not between -360 and 360")
    else:
        if np.abs(lam_r) > (2 * np.pi):
            raise ValueError(
                "rotate: lam_r ({} deg) is not between -360 and 360".format(
                    180 * lam_r / np.pi))

    # difference between azimuth and rotation angle of x-axis
    omega = lam_r - azm

    # sine of angle theta (cosine is an argument)
    # sine of angle theta-sub-r (cosine is an argument)
    sinTheta = np.sqrt((1. - mu) * (1. + mu))
    sinThr = np.sqrt((1. - mu_r) * (1. + mu_r))

    # cosine of difference between azimuth and aziumth of rotation
    cosOmega = np.cos(omega)

    # Output:
    # azimuth and cosine of angle in rotated axis system.
    # (bug if trig routines trigger error)
    aPrime = -np.arctan2(sinTheta * np.sin(omega),
                         mu_r * sinTheta * cosOmega - mu * sinThr)
    muPrime = sinTheta * sinThr * cosOmega + mu * mu_r

    return muPrime, aPrime


def yearday(year, month, day):
    """
    yearday returns the yearday for the given date.  yearday is
    the 'day of the year', sometimes called (incorrectly) 'julian day'.

    Args:
        year
        month
        day

    Returns:
        yday: day of year

    """

    assert(year >= 0)
    assert(1 <= month and month <= 12)
    assert(1 <= day and day <= numdays(year, month))

    yday = day

    # Add in number of days for preceeding months.
    for mon in range(month-1, 0, -1):
        yday += numdays(year, mon)

    return yday


def numdays(year, month):
    """
    numdays returns the number of days in the given month of
    the given year.

    Args:
        year
        month

    Returns:
        ndays: number of days in month
    """

    NDAYS = list([31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31])

    assert(year >= 0)
    assert(1 <= month and month <= 12)

    ndays = NDAYS[month-1]

    # Check for leap year for February
    if ((month == 2) and leapyear(year)):
        ndays += 1

    return ndays


def leapyear(year):
    """
    leapyear determines if the given year is a leap year or not.
    year must be positive, and must not be abbreviated; i.e.
    89 is 89 A.D. not 1989.

    Args:
        year

    Returns:
        True if a leap year, False if not a leap year
    """

    assert(year >= 0)

    if ((year % 4 == 0 and year % 100 != 0) or year % 400 == 0):
        return True
    else:
        return False



import os
from copy import copy
from glob import glob

from inicheck.tools import cast_all_variables

from smrf.framework.model_framework import run_smrf
from smrf.tests.smrf_test_case import SMRFTestCase


class TestOutputThreadedVariables(SMRFTestCase):

    def tearDown(self):
        self.remove_output_dir()

    def change_variables(self, new_variables):
        config = copy(self.base_config)
        config.raw_cfg['output'].update({
            'variables': new_variables
        })

        config.apply_recipes()
        config = cast_all_variables(config, config.mcfg)

        return config

    def check_outputs(self, new_variables):

        files = glob(os.path.join(
            self.base_config.cfg['output']['out_location'], '*.nc'))
        out_files = [os.path.basename(f).split('.')[0] for f in files]

        out_files.sort()
        new_variables.sort()

        self.assertTrue(new_variables == out_files)

    def test_variables_standard(self):

        new_variables = ['thermal', 'air_temp']
        config = self.change_variables(new_variables)
        run_smrf(config)
        self.check_outputs(new_variables)

    def test_variables_non_standard(self):

        new_variables = ['clear_ir_beam', 'veg_vis_diffuse',
                         'cloud_vis_beam', 'thermal_veg']
        config = self.change_variables(new_variables)
        run_smrf(config)
        self.check_outputs(new_variables)


class TestOutputVariables(TestOutputThreadedVariables):

    @classmethod
    def setUpClass(cls):
        """
        Runs the short simulation over reynolds mountain east
        """
        super().setUpClass()

        config = cls.base_config_copy()
        config.raw_cfg['system']['threading'] = False

        config.apply_recipes()
        config = cast_all_variables(config, config.mcfg)
        cls._base_config = config



# cython: embedsignature=True
# cython: language_level=3
"""
C implementation of some radiation functions
"""


import cython
import numpy as np

cimport numpy as np

# Numpy must be initialized. When using numpy from C or Cython you must
# _always_ do that, or you will have segfaults
np.import_array()


cdef extern from "envphys_c.h":
    void topotherm(int ngrid, double *ta, double *tw, double *z, double *skvfac, int nthreads, double *thermal);
    void dewpt(int ngrid, double *ea, int nthreads, double tol, double *dpt);
    void iwbt(int ngrid, double *ta, double *td,	double *z, int nthreads, double tol, double *tw);


@cython.boundscheck(False)
@cython.wraparound(False)
# https://github.com/cython/cython/wiki/tutorials-NumpyPointerToC
def ctopotherm(np.ndarray[double, mode="c", ndim=2] ta,
               np.ndarray[double, mode="c", ndim=2] tw,
               np.ndarray[double, mode="c", ndim=2] z,
               np.ndarray[double, mode="c", ndim=2] skvfac,
               np.ndarray[double, mode="c", ndim=2] thermal not None,
               int nthreads=1):
    '''
    Call the function krige_grid in krige.c which will iterate over the grid
    within the C code

    Args:
        ta, tw, z, skvfac
    Out:
        thermal changed in place

    20160325 Scott Havens
    '''

    cdef int ngrid
    ngrid = ta.shape[0] * ta.shape[1]

    # convert the ta array to C
    cdef np.ndarray[double, mode="c", ndim=2] ta_arr
    ta_arr = np.ascontiguousarray(ta, dtype=np.float64)

    # convert the tw array to C
    cdef np.ndarray[double, mode="c", ndim=2] tw_arr
    tw_arr = np.ascontiguousarray(tw, dtype=np.float64)

    # convert the ta array to C
    cdef np.ndarray[double, mode="c", ndim=2] z_arr
    z_arr = np.ascontiguousarray(z, dtype=np.float64)

    # convert the skvfac to C
    cdef np.ndarray[double, mode="c", ndim=2] skvfac_arr
    skvfac_arr = np.ascontiguousarray(skvfac, dtype=np.float64)

    # call the C function
    topotherm(ngrid, &ta_arr[0,0], &tw_arr[0,0], &z_arr[0,0], &skvfac_arr[0,0], nthreads, &thermal[0,0])

    return None


@cython.boundscheck(False)
@cython.wraparound(False)
# https://github.com/cython/cython/wiki/tutorials-NumpyPointerToC
def cdewpt(np.ndarray[double, mode="c", ndim=2] vp,
               np.ndarray[double, mode="c", ndim=2] dwpt not None,
               float tolerance=0,
               int nthreads=1):
    '''

    Args:
        vp
    Out:
        dwpt changed in place

    20160505 Scott Havens
    '''

    cdef int ngrid
    ngrid = vp.shape[0] * vp.shape[1]

    # convert the ta array to C
    cdef np.ndarray[double, mode="c", ndim=2] vp_arr
    vp_arr = np.ascontiguousarray(vp, dtype=np.float64)

    # call the C function
    dewpt(ngrid, &vp_arr[0,0], nthreads, tolerance, &dwpt[0,0])

    return None


@cython.boundscheck(False)
@cython.wraparound(False)
# https://github.com/cython/cython/wiki/tutorials-NumpyPointerToC
def cwbt(np.ndarray[double, mode="c", ndim=2] ta,
         np.ndarray[double, mode="c", ndim=2] td,
         np.ndarray[double, mode="c", ndim=2] z,
         np.ndarray[double, mode="c", ndim=2] tw not None,
         float tolerance=0,
         int nthreads=1):
    '''
    Call the function iwbt in iwbt.c which will iterate over the grid
    within the C code

    Args:
        ta, td, z
    Out:
        tw changed in place (wet bulb temperature)

    20180611 Micah Sandusky
    '''

    cdef int ngrid
    ngrid = ta.shape[0] * ta.shape[1]

    # convert the ta array to C
    cdef np.ndarray[double, mode="c", ndim=2] ta_arr
    ta_arr = np.ascontiguousarray(ta, dtype=np.float64)

    # convert the dew point array to C
    cdef np.ndarray[double, mode="c", ndim=2] td_arr
    td_arr = np.ascontiguousarray(td, dtype=np.float64)

    # convert the ta array to C
    cdef np.ndarray[double, mode="c", ndim=2] z_arr
    z_arr = np.ascontiguousarray(z, dtype=np.float64)

    # call the C function
    iwbt(ngrid, &ta_arr[0,0], &td_arr[0,0], &z_arr[0,0], nthreads, tolerance, &tw[0,0])

    return None




import numpy as np

from smrf.distribute import image_data
from smrf.envphys.core import envphys_c
from smrf.utils import utils


class vp(image_data.image_data):
    """
    The :mod:`~smrf.distribute.vapor_pressure.vp` class allows for variable
    specific distributions that go beyond the base class

    Vapor pressure is provided as an argument and is calculated from coincident
    air temperature and relative humidity measurements using utilities such as
    :mod:`smrf.envphys.vapor_pressure.rh2vp`. The vapor pressure is distributed
    instead of the relative humidity as it is an absolute measurement of the
    vapor within the atmosphere and will follow elevational trends (typically
    negative).  Were as relative humidity is a relative measurement which
    varies in complex ways over the topography.  From the distributed vapor
    pressure, the dew point is calculated for use by other distribution
    methods. The dew point temperature is further corrected to ensure that it
    does not exceed the distributed air temperature.

    Args:
        vpConfig: The [vapor_pressure] section of the configuration file

    Attributes:
        config: configuration from [vapor_pressure] section
        vapor_pressure: numpy matrix of the vapor pressure
        dew_point: numpy matrix of the dew point, calculated from
            vapor_pressure and corrected for dew_point greater than air_temp
        min: minimum value of vapor pressure is 10 Pa
        max: maximum value of vapor pressure is 7500 Pa
        stations: stations to be used in alphabetical order

    """

    variable = 'vapor_pressure'

    # these are variables that can be output
    OUTPUT_VARIABLES = {
        'vapor_pressure': {
            'units': 'pascal',
            'standard_name': 'vapor_pressure',
            'long_name': 'Vapor pressure'
        },
        'dew_point': {
            'units': 'degree_Celcius',
            'standard_name': 'dew_point_temperature',
            'long_name': 'Dew point temperature'
        },
        'precip_temp': {
            'units': 'degree_Celcius',
            'standard_name': 'precip_temperature',
            'long_name': 'Precip temperature'
        }
    }

    # these are variables that are operate at the end only and do not need to
    # be written during main distribute loop
    post_process_variables = {}

    BASE_THREAD_VARIABLES = frozenset([
        'vapor_pressure',
        'dew_point',
        'precip_temp'
    ])

    def __init__(self, vpConfig, precip_temp_method):

        # extend the base class
        image_data.image_data.__init__(self, self.variable)

        # check and assign the configuration
        self.getConfig(vpConfig)

        # assign precip temp method
        self.precip_temp_method = precip_temp_method

        self._logger.debug('Created distribute.vapor_pressure')

    def initialize(self, topo, data, date_time=None):
        """
        Initialize the distribution, calls
        :mod:`smrf.distribute.image_data.image_data._initialize`. Preallocates
        the following class attributes to zeros:

        Args:
            topo: :mod:`smrf.data.loadTopo.Topo` instance contain topographic
                data and infomation
            data: data Pandas dataframe containing the station data,
                from :mod:`smrf.data.loadData` or :mod:`smrf.data.loadGrid`

        """

        self._logger.debug('Initializing distribute.vapor_pressure')
        self.date_time = date_time
        self._initialize(topo, data.metadata)

        # get dem to pass to wet_bulb
        self.dem = topo.dem

    def distribute(self, data, ta):
        """
        Distribute air temperature given a Panda's dataframe for a single time
        step. Calls :mod:`smrf.distribute.image_data.image_data._distribute`.

        The following steps are performed when distributing vapor pressure:

        1. Distribute the point vapor pressure measurements
        2. Calculate dew point temperature using
            :mod:`smrf.envphys.core.envphys_c.cdewpt`
        3. Adjust dew point values to not exceed the air temperature

        Args:
            data: Pandas dataframe for a single time step from precip
            ta: air temperature numpy array that will be used for calculating
                dew point temperature

        """

        self._logger.debug('%s -- Distributing vapor_pressure' % data.name)

        # calculate the vapor pressure
        self._distribute(data)

        # set the limits
        self.vapor_pressure = utils.set_min_max(self.vapor_pressure,
                                                self.min,
                                                self.max)

        # calculate the dew point
        self._logger.debug('%s -- Calculating dew point' % data.name)

        # use the core_c to calculate the dew point
        dpt = np.zeros_like(self.vapor_pressure, dtype=np.float64)
        envphys_c.cdewpt(self.vapor_pressure,
                         dpt,
                         self.config['dew_point_tolerance'],
                         self.config['dew_point_nthreads'])

        # find where dpt > ta
        ind = dpt >= ta

        if (np.sum(ind) > 0):  # or np.sum(indm) > 0):
            dpt[ind] = ta[ind] - 0.2

        self.dew_point = dpt

        # calculate wet bulb temperature
        if self.precip_temp_method == 'wet_bulb':
            # initialize timestep wet_bulb
            wet_bulb = np.zeros_like(self.vapor_pressure, dtype=np.float64)
            # calculate wet_bulb
            envphys_c.cwbt(ta, dpt, self.dem,
                           wet_bulb, self.config['dew_point_tolerance'],
                           self.config['dew_point_nthreads'])
            # # store last time step of wet_bulb
            # self.wet_bulb_old = wet_bulb.copy()
            # store in precip temp for use in precip
            self.precip_temp = wet_bulb
        else:
            self.precip_temp = dpt

    def distribute_thread(self, smrf_queue, data_queue):
        """
        Distribute the data using threading. All data is provided and
        ``distribute_thread`` will go through each time step and call
        :mod:`smrf.distribute.vapor_pressure.vp.distribute` then puts the
        distributed data into the smrf_queue for:

        * :py:attr:`vapor_pressure`
        * :py:attr:`dew_point`

        Args:
            smrf_queue: smrf_queue dictionary for all variables
            data: pandas dataframe for all data, indexed by date time
        """
        self._logger.info("Distributing {}".format(self.variable))

        for date_time in self.date_time:

            vp_data = data_queue['vapor_pressure'].get(date_time)
            ta = smrf_queue['air_temp'].get(date_time)

            self.distribute(vp_data, ta)

            smrf_queue[self.variable].put([date_time, self.vapor_pressure])
            smrf_queue['precip_temp'].put([date_time, self.precip_temp])
            smrf_queue['dew_point'].put([date_time, self.dew_point])



import logging

import numpy as np

from smrf.spatial import dk, grid, idw, kriging


class image_data():
    """
    A base distribution method in SMRF that will ensure all variables are
    distributed in the same manner.  Other classes will be initialized
    using this base class.

    .. code-block:: Python

        class ta(smrf.distribute.image_data):
            '''
            This is the ta class extending the image_data base class
            '''

    Args:
        variable (str): Variable name for the class

    Returns:
        A :mod:`!smrf.distribute.image_data` class instance

    Attributes:
        variable: The name of the variable that this class will become
        [variable_name]: The :py:attr:`variable` will have the distributed data
        [other_attribute]: The distributed data can also be stored as another
            attribute specified in
            :mod:`~smrf.distribute.image_data._distribute`
        config: Parsed dictionary from the configuration file for the variable
        stations: The stations to be used for the variable, if set, in
            alphabetical order
        metadata: The metadata Pandas dataframe containing the station
            information from :mod:`smrf.data.loadData` or
            :mod:`smrf.data.loadGrid`
        idw: Inverse distance weighting instance from
            :mod:`smrf.spatial.idw.IDW`
        dk: Detrended kriging instance from :mod:`smrf.spatial.dk.dk.DK`
        grid: Gridded interpolation instance from :mod:`smrf.spatial.grid.GRID`

    """

    BASE_THREAD_VARIABLES = frozenset()
    OUTPUT_VARIABLES = {}

    def __init__(self, variable):

        self.variable = variable
        setattr(self, variable, None)

        self.gridded = False

        self._logger = logging.getLogger(self.__class__.__module__)
        self._thread_variables = None

    @property
    def thread_variables(self):
        if self._thread_variables is None:
            self._thread_variables = list(self.BASE_THREAD_VARIABLES)

        return self._thread_variables

    @thread_variables.setter
    def thread_variables(self, value):
        self._thread_variables = value

    def add_thread_variables(self, variables):
        """Add a list or single variable to the thread variables

        Args:
            variables (list or str): List or string of variables to add
        """

        if isinstance(variables, str):
            variables = list((variables, ))

        self.thread_variables = self.thread_variables + variables

    @property
    def output_variables(self):
        ov = {}
        for key, value in self.OUTPUT_VARIABLES.items():
            value['module'] = self.__class__.__module__.split('.')[-1]
            ov[key] = value
        return ov

    def getConfig(self, cfg):
        """
        Check the configuration that was set by the user for the variable
        that extended this class. Checks for standard distribution parameters
        that are common across all variables and assigns to the class instance.
        Sets the :py:attr:`config` and :py:attr:`stations` attributes.

        Args:
            cfg (dict): dict from the [variable]
        """

        # check of gridded interpolation
        self.gridded = False
        if 'distribution' in cfg.keys():
            if cfg['distribution'] == 'grid':
                self.gridded = True

        self.getStations(cfg)
        self.config = cfg

    def getStations(self, config):
        """
        Determines the stations from the [variable] section of the
        configuration file.

        Args:
            config (dict): dict from the [variable]
        """
        stations = None

        # determine the stations that will be used, alphabetical order
        if "stations" in config.keys():
            if config["stations"] is not None:
                stations = config['stations'].sort()

        self.stations = stations

    def _initialize(self, topo, metadata):
        """
        Initialize the distribution based on the parameters in
        :py:attr:`config`.

        Args:
            topo: :mod:`smrf.data.loadTopo.Topo` instance contain topographic
                data and infomation
            metadata: metadata Pandas dataframe containing the station metadata
                from :mod:`smrf.data.loadData` or :mod:`smrf.data.loadGrid`

        Raises:
            Exception: If the distribution method could not be determined, must
                be idw, dk, or grid

        To do:
            - make a single call to the distribution initialization
            - each dist (idw, dk, grid) takes the same inputs and returns the
                same
        """

        self.min = self.config['min']
        if self.min is None:
            self.min = -np.inf

        self.max = self.config['max']
        if self.max is None:
            self.max = np.inf

        # pull out the metadata subset
        if self.stations is not None:
            metadata = metadata.loc[self.stations]
        else:
            self.stations = metadata.index.values
        self.metadata = metadata

        # Old DB used X and Y, New DB uses utm_x, utm_y
        try:
            self.mx = metadata.utm_x.values
            self.my = metadata.utm_y.values
        except Exception:
            self.mx = metadata.X.values
            self.my = metadata.Y.values

        self.mz = metadata.elevation.values

        if "distribution" in self.config.keys():
            if self.config['distribution'] == 'idw':
                # inverse distance weighting
                self.idw = idw.IDW(
                    self.mx, self.my, topo.X, topo.Y, mz=self.mz,
                    GridZ=topo.dem, power=self.config['idw_power'])

            elif self.config['distribution'] == 'dk':
                # detrended kriging
                self.dk = dk.DK(
                    self.mx, self.my, self.mz, topo.X, topo.Y,
                    topo.dem,
                    self.config)

            elif self.config['distribution'] == 'grid':
                # linear interpolation between points
                self.grid = grid.GRID(
                    self.config, self.mx, self.my, topo.X,
                    topo.Y,
                    mz=self.mz,
                    GridZ=topo.dem,
                    mask=topo.mask,
                    metadata=metadata)

            elif self.config['distribution'] == 'kriging':
                # generic kriging
                self.kriging = kriging.KRIGE(
                    self.mx, self.my, self.mz, topo.X,
                    topo.Y,
                    topo.dem,
                    self.config)

            else:
                raise Exception(
                    "Could not determine the distribution method for "
                    "{}".format(self.variable))

    def _distribute(self, data, other_attribute=None, zeros=None):
        """
        Distribute the data using the defined distribution method in
        :py:attr:`config`

        Args:
            data: Pandas dataframe for a single time step
            other_attribute (str): By default, the distributed data matrix goes
                into self.variable but this specifies another attribute in self
            zeros: data values that should be treated as zeros (not used)

        Raises:
            Exception: If all input data is NaN
        """

        # get the data for the desired stations
        # this will also order it correctly how air_temp was initialized
        data = data[self.stations]

        if np.sum(data.isnull()) == data.shape[0]:
            raise Exception("{}: All data values are NaN"
                            "".format(self.variable))

        if self.config['distribution'] == 'idw':
            if self.config['detrend']:
                v = self.idw.detrendedIDW(
                    data.values,
                    self.config['detrend_slope'],
                    zeros=zeros)
            else:
                v = self.idw.calculateIDW(data.values)

        elif self.config['distribution'] == 'dk':
            v = self.dk.calculate(data.values)

        elif self.config['distribution'] == 'grid':
            if self.config['detrend']:
                v = self.grid.detrendedInterpolation(
                    data,
                    self.config['detrend_slope'],
                    self.config['grid_method'])
            else:
                v = self.grid.calculateInterpolation(
                    data.values,
                    self.config['grid_method'])

        elif self.config['distribution'] == 'kriging':
            v, ss = self.kriging.calculate(data.values)
            setattr(self, '{}_variance'.format(self.variable), ss)

        if other_attribute is not None:
            setattr(self, other_attribute, v)
        else:
            setattr(self, self.variable, v)

    def post_processor(self, output_func):
        """
        Each distributed variable has the oppurtunity to do post processing on
        a sub variable. This is necessary in cases where the post proecessing
        might need to be done on a different timescale than that of the main
        loop.

        Should be redefined in the individual variable module.
        """
        pass



from smrf.envphys.nasde_model import Susong1999
import unittest
import numpy as np


class TestSusong1999(unittest.TestCase):
    TEMPERATURES = np.array([-5, -3, 0])
    PRECIPITATION = np.array([1, 1, 1])

    def test_returns_precipitation(self):
        results = Susong1999.run(self.TEMPERATURES, self.PRECIPITATION)

        np.testing.assert_equal(results['rho_s'], [100, 150, 250])
        assert 'pcs' in results

    def test_returns_no_precipitation(self):
        precipitation = np.zeros(self.TEMPERATURES.shape)

        results = Susong1999.run(self.TEMPERATURES, precipitation)

        assert 'rho_s' in results
        assert 'pcs' in results



import numpy as np

from smrf.envphys.constants import EMISS_TERRAIN, FREEZE, STEF_BOLTZ


def thermal_correct_terrain(th, ta, viewf):
    """
    Correct the thermal radiation for terrain assuming that
    the terrain is at the air temperature and the pixel and
    a sky view

    Args:
        th: thermal radiation
        ta: air temperature [C]
        viewf: sky view factor from view_f

    Returns:
        corrected thermal radiation

    20150611 Scott Havens
    """

    # thermal emitted from the terrain
    terrain = STEF_BOLTZ * EMISS_TERRAIN * np.power(ta + 273.15, 4)

    # correct the incoming thermal
    return viewf * th + (1 - viewf) * terrain


def precipitable_water(ta, ea):
    """
    Estimate the precipitable water from Prata (1996) :cite:`Prata:1996`
    """
    return 4650*ea/ta


def calc_long_wave(e, ta):
    """
    Apply the Stephan-Boltzman equation for longwave
    """
    return e * STEF_BOLTZ * np.power(ta, 4)


def Dilly1998(ta, ea):
    """
    Estimate clear-sky downwelling long wave radiation from Dilley & O'Brian
    (1998) :cite:`Dilley&OBrian:1998` using the equation:

    .. math::
        L_{clear} = 59.38 + 113.7 * \\left( \\frac{T_a}{273.16}
        \\right)^6 + 96.96 \\sqrt{w/25}

    Where :math:`T_a` is the air temperature and :math:`w` is the amount of
    precipitable water. The preipitable water is estimated as
    :math:`4650 e_o/T_o` from Prata (1996) :cite:`Prata:1996`.

    Args:
        ta: distributed air temperature [degree C]
        ea: distrubted vapor pressure [kPa]

    Returns:
        clear sky long wave radiation [W/m2]

    20170509 Scott Havens
    """

    ta = ta + FREEZE                    # convert to K
    w = precipitable_water(ta, ea)  # precipitable water

    th = 59.38 + 113.7 * (ta/273.16)**6 + 96.96*np.sqrt(w/25)

    return th


def Prata1996(ta, ea):
    """
    Estimate clear-sky downwelling long wave radiation from Prata (1996)
    :cite:`Prata:1996` using the equation:

    .. math::
        \\epsilon_{clear} = 1 - (1 + w) * exp(-1.2 + 3w)^{1/2}

    Where :math:`w` is the amount of precipitable water. The preipitable
    water is estimated as :math:`4650 e_o/T_o` from Prata (1996)
    :cite:`Prata:1996`.

    Args:
        ta: distributed air temperature [degree C]
        ea: distrubted vapor pressure [kPa]

    Returns:
        clear sky long wave radiation [W/m2]

    20170509 Scott Havens
    """

    ta = ta + FREEZE                    # convert to K
    w = precipitable_water(ta, ea)  # precipitable water

    # clear sky emmissivity
    ec = 1 - (1 + w) * np.exp(-np.sqrt(1.2 + 3*w))

    return calc_long_wave(ec, ta)


def Angstrom1918(ta, ea):
    """
    Estimate clear-sky downwelling long wave radiation from Angstrom (1918)
    :cite:`Angstrom:1918` as cited by Niemela et al (2001)
    :cite:`Niemela&al:2001` using the equation:

    .. math::
        \\epsilon_{clear} = 0.83 - 0.18 * 10^{-0.067 e_a}

    Where :math:`e_a` is the vapor pressure.

    Args:
        ta: distributed air temperature [degree C]
        ea: distrubted vapor pressure [kPa]

    Returns:
        clear sky long wave radiation [W/m2]

    20170509 Scott Havens
    """

    ta = ta + FREEZE                  # convert to K
    e = 0.83 - 0.18 * np.power(10, -0.067*ea)

    return calc_long_wave(e, ta)



import numpy as np

from smrf.distribute import image_data
from smrf.envphys import albedo
from smrf.utils import utils


class Albedo(image_data.image_data):
    """
    The :mod:`~smrf.distribute.albedo.Albedo` class allows for variable
    specific distributions that go beyond the base class.

    The visible (280-700nm) and infrared (700-2800nm) albedo follows the
    relationships described in Marks et al. (1992) :cite:`Marks&al:1992`. The
    albedo is a function of the time since last storm, the solar zenith angle,
    and grain size. The time since last storm is tracked on a pixel by pixel
    basis and is based on where there is significant accumulated distributed
    precipitation. This allows for storms to only affect a small part of the
    basin and have the albedo decay at different rates for each pixel.

    Args:
        albedoConfig: The [albedo] section of the configuration file

    Attributes:
        albedo_vis: numpy array of the visible albedo
        albedo_ir: numpy array of the infrared albedo
        config: configuration from [albedo] section
        min: minimum value of albedo is 0
        max: maximum value of albedo is 1
        stations: stations to be used in alphabetical order
    """

    variable = 'albedo'

    # these are variables that can be output
    OUTPUT_VARIABLES = {
        'albedo_vis': {
            'units': 'None',
            'standard_name': 'visible_albedo',
            'long_name': 'Visible wavelength albedo'
        },
        'albedo_ir': {
            'units': 'None',
            'standard_name': 'infrared_albedo',
            'long_name': 'Infrared wavelength albedo'
        }
    }
    # these are variables that are operate at the end only and do not need to
    # be written during main distribute loop
    post_process_variables = {}

    BASE_THREAD_VARIABLES = frozenset([
        'albedo_vis',
        'albedo_ir'
    ])

    def __init__(self, albedoConfig):
        """
        Initialize albedo()

        Args:
            albedoConfig: configuration from [albedo] section
        """

        # extend the base class
        image_data.image_data.__init__(self, self.variable)

        # Get the veg values for the decay methods. Date method uses self.veg
        # Hardy2000 uses self.litter
        for d in ['veg', 'litter']:
            v = {}

            matching = [s for s in albedoConfig.keys()
                        if "{0}_".format(d) in s]
            for m in matching:
                ms = m.split('_')
                v[ms[-1]] = albedoConfig[m]

            # Create self.litter,self.veg
            setattr(self, d, v)

        self.getConfig(albedoConfig)

        self._logger.debug('Created distribute.albedo')

    def initialize(self, topo, data, date_time=None):
        """
        Initialize the distribution, calls image_data.image_data._initialize()

        Args:
            topo: smrf.data.loadTopo.Topo instance contain topo data/info
            data: data dataframe containing the station data

        """

        self._logger.debug('Initializing distribute.albedo')
        self.veg_type = topo.veg_type
        self.date_time = date_time
        self._initialize(topo, data.metadata)

        if self.config["decay_method"] is None:
            self._logger.warning("No decay method is set!")

    def distribute(self, current_time_step, cosz, storm_day):
        """
        Distribute air temperature given a Panda's dataframe for a single time
        step. Calls :mod:`smrf.distribute.image_data.image_data._distribute`.

        Args:
            current_time_step: Current time step in datetime object
            cosz: numpy array of the illumination angle for the current time
                step
            storm_day: numpy array of the decimal days since it last
                snowed at a grid cell

        """

        self._logger.debug('%s Distributing albedo' % current_time_step)

        # only need to calculate albedo if the sun is up
        if cosz is not None:

            alb_v, alb_ir = albedo.albedo(
                storm_day, cosz,
                self.config['grain_size'],
                self.config['max_grain'],
                self.config['dirt'])

            # Perform litter decay
            if self.config['decay_method'] == 'date_method':
                alb_v_d, alb_ir_d = albedo.decay_alb_power(
                    self.veg,
                    self.veg_type,
                    self.config['date_method_start_decay'],
                    self.config['date_method_end_decay'],
                    current_time_step,
                    self.config['date_method_decay_power'],
                    alb_v, alb_ir)

                alb_v = alb_v_d
                alb_ir = alb_ir_d

            elif self.config['decay_method'] == 'hardy2000':
                alb_v_d, alb_ir_d = albedo.decay_alb_hardy(
                    self.litter,
                    self.veg_type,
                    storm_day,
                    alb_v,
                    alb_ir)

                alb_v = alb_v_d
                alb_ir = alb_ir_d

            self.albedo_vis = utils.set_min_max(alb_v, self.min, self.max)
            self.albedo_ir = utils.set_min_max(alb_ir, self.min, self.max)

        else:
            self.albedo_vis = np.zeros(storm_day.shape)
            self.albedo_ir = np.zeros(storm_day.shape)

    def distribute_thread(self, smrf_queue, data_queue=None):
        """
        Distribute the data using threading and queue

        Args:
            queue: queue dict for all variables
            date: dates to loop over

        Output:
            Changes the queue albedo_vis, albedo_ir
                for the given date
        """
        self._logger.info("Distributing {}".format(self.variable))

        for date_time in self.date_time:

            illum_ang = smrf_queue['illum_ang'].get(date_time)
            storm_day = smrf_queue['storm_days'].get(date_time)

            self.distribute(date_time, illum_ang, storm_day)

            smrf_queue['albedo_vis'].put([date_time, self.albedo_vis])
            smrf_queue['albedo_ir'].put([date_time, self.albedo_ir])



import logging

import numpy as np
import utm

from smrf.data import InputCSV, InputGribHRRR, InputNetcdf, InputWRF


class InputData():
    """
    Class for loading and storing the data, either from
    - CSV file
    - HRRR grib files
    - WRF
    - Generic gridded NetCDF


    The data will be loaded into a Pandas dataframe

    """

    VARIABLES = [
        'air_temp',
        'vapor_pressure',
        'precip',
        'wind_speed',
        'wind_direction',
        'cloud_factor',
        'thermal',
        'metadata'
    ]

    # Data variables and which module they belong to
    MODULE_VARIABLES = {
        'air_temp': 'air_temp',
        'vapor_pressure': 'vapor_pressure',
        'precip': 'precipitation',
        'wind_speed': 'wind',
        'wind_direction': 'wind',
        'cloud_factor': 'cloud_factor',
        'thermal': 'thermal'
    }

    DATA_FUNCTIONS = {
        InputCSV.DATA_TYPE: InputCSV,
        InputWRF.DATA_TYPE: InputWRF,
        InputNetcdf.DATA_TYPE: InputNetcdf,
        InputGribHRRR.DATA_TYPE: InputGribHRRR
    }

    # degree offset for a buffer around the model domain in degrees
    OFFSET = 0.1

    def __init__(self, smrf_config, start_date, end_date, topo):

        self.smrf_config = smrf_config
        self.start_date = start_date
        self.end_date = end_date
        self.time_zone = start_date.tzinfo
        self.topo = topo

        self._logger = logging.getLogger(__name__)

        # get the buffer gridded data domain extents in lat long
        self.model_domain_grid()

        if 'csv' in self.smrf_config:
            self.data_type = InputCSV.DATA_TYPE
            data_inputs = {
                'stations': self.smrf_config['csv']['stations'],
                'config': self.smrf_config['csv']
            }

        elif 'gridded' in self.smrf_config:
            self.data_type = self.smrf_config['gridded']['data_type']
            data_inputs = {
                'bbox': self.bbox,
                'config': self.smrf_config['gridded'],
                'topo': self.topo
            }

        # load the data
        self.load_class = self.DATA_FUNCTIONS[self.data_type](
            self.start_date,
            self.end_date,
            **data_inputs)
        self.load_class.load()

        self.set_variables()

        self.metadata_pixel_location()

        if self.data_type == InputCSV.DATA_TYPE:
            self.load_class.check_colocation()

    def set_variables(self):
        """Set the instance attributes for each variable
        """

        for variable in self.VARIABLES:
            d = getattr(self.load_class, variable, None)
            if variable == 'metadata':
                setattr(self, variable, d)
            elif d is not None:
                d = d.tz_convert(tz=self.time_zone)
                setattr(self, variable, d[self.start_date:self.end_date])

    def metadata_pixel_location(self):
        """Set the pixel location in the topo for each station
        """

        self.metadata['xi'] = self.metadata.apply(
            lambda row: self.find_pixel_location(
                row,
                self.topo.x,
                'utm_x'), axis=1)
        self.metadata['yi'] = self.metadata.apply(
            lambda row: self.find_pixel_location(
                row,
                self.topo.y,
                'utm_y'), axis=1)

    def model_domain_grid(self):
        '''
        Retrieve the bounding box for the gridded data by adding a buffer to
        the extents of the topo domain.

        Returns:
            tuple: (dlat, dlon) Domain latitude and longitude extents
        '''
        dlat = np.zeros((2,))
        dlon = np.zeros_like(dlat)

        # Convert the UTM extents to lat long extents
        ur = self.get_latlon(np.max(self.topo.x), np.max(self.topo.y))
        ll = self.get_latlon(np.min(self.topo.x), np.min(self.topo.y))

        # Put into numpy arrays for convenience later
        dlat[0], dlon[0] = ll
        dlat[1], dlon[1] = ur

        # add a buffer
        dlat[0] -= self.OFFSET
        dlat[1] += self.OFFSET
        dlon[0] -= self.OFFSET
        dlon[1] += self.OFFSET

        self.dlat = dlat
        self.dlon = dlon

        # This is placed long, lat on purpose as thats the HRRR class expects
        self.bbox = np.array([self.dlon[0], self.dlat[0],
                              self.dlon[1], self.dlat[1]])

    def get_latlon(self, utm_x, utm_y):
        '''
        Convert UTM coords to Latitude and longitude

        Args:
            utm_x: UTM easting in meters in the same zone/letter as the topo
            utm_y: UTM Northing in meters in the same zone/letter as the topo

        Returns:
            tuple: (lat,lon) latitude and longitude conversion from the UTM
                coordinates
        '''

        lat, lon = utm.to_latlon(utm_x, utm_y, self.topo.zone_number,
                                 northern=self.topo.northern_hemisphere)
        return lat, lon

    def find_pixel_location(self, row, vec, a):
        """
        Find the index of the stations X/Y location in the model domain

        Args:
            row (pandas.DataFrame): metadata rows
            vec (nparray): Array of X or Y locations in domain
            a (str): Column in DataFrame to pull data from (i.e. 'X')

        Returns:
            Pixel value in vec where row[a] is located
        """
        return np.argmin(np.abs(vec - row[a]))



import numpy as np

from smrf.envphys.constants import (FREEZE, GRAVITY, MOL_AIR, RGAS, SEA_LEVEL,
                                    STD_LAPSE, STD_LAPSE_M, STEF_BOLTZ)
from smrf.envphys.vapor_pressure import sati


def brutsaert(air_temp, lapse_rate, vapor_pressure, elevation, pressure):
    """
    Calculate atmosphere emissivity from Brutsaert (1975):cite:`Brutsaert:1975`

    Args:
        air_temp: air temp (K)
        lapse_rate: temperature lapse rate (deg/m)
        ea: vapor pressure (Pa)
        elevation: elevation (z)
        pressure: air pressure (Pa)

    Returns:
        atmosphericy emissivity

    20151027 Scott Havens
    """

    t_prime = air_temp - (lapse_rate * elevation)
    rh = vapor_pressure / sati(air_temp)
    rh[rh > 1] = 1

    e_prime = (rh * sati(t_prime)) / 100.0

    air_emiss = (1.24 * np.power(e_prime / t_prime, 1. / 7.0)) * \
        pressure / SEA_LEVEL

    air_emiss[air_emiss > 1.0] = 1.0

    return air_emiss


def hysat(pb, tb, L, h, g, m):
    """
    integral of hydrostatic equation over layer with linear temperature
    variation

    Args:
        pb: base level pressure
        tb: base level temp [K]
        L: lapse rate [deg/km]
        h: layer thickness [km]
        g: grav accel [m/s^2]
        m: molec wt [kg/kmole]

    Returns:
        hydrostatic results

    20151027 Scott Havens
     """

    # the factors 1.e-3 and 1.e3 are for units conversion
    if L == 0:
        return pb * np.exp(-g * m * h * 1.e3/(RGAS * tb))
    else:
        return pb * np.power(tb/(tb + L * h), g * m/(RGAS * L * 1.e-3))


def topotherm(ta, tw, z, skvfac):
    """
    Calculate the clear sky thermal radiation.  topotherm calculates  thermal
    radiation from the atmosphere corrected for topographic effects, from near
    surface air temperature Ta, dew point temperature DPT, and elevation. Based
    on a model by Marks and Dozier (1979) :citeL`Marks&Dozier:1979`.

    Args:
        ta: air temperature [C]
        tw: dew point temperature [C]
        z: elevation [m]
        skvfac: sky view factor

    Returns:
        Long wave (thermal) radiation corrected for terrain

    20151027 Scott Havens
    """

    # convert ta and tw from C to K
    ta = ta + FREEZE
    tw = tw + FREEZE

    # if below zero set to nan
    tw[tw < 0] = np.nan
    ta[ta < 0] = np.nan

    # calculate theoretical sea level
    # atmospheric emissivity
    # from reference level ta, tw, and z
    ind = tw > ta
    tw[ind] = ta[ind]

    ea = sati(tw)
    emiss = brutsaert(ta, STD_LAPSE_M, ea, z, SEA_LEVEL)

    # calculate sea level air temp
    T0 = ta - (z * STD_LAPSE_M)

    # adjust emiss for elev, terrain
    # veg, and cloud shading
    press = hysat(SEA_LEVEL, T0, STD_LAPSE, z/1000.0, GRAVITY, MOL_AIR)

    # elevation correction
    emiss *= press/SEA_LEVEL

    # terrain factor correction
    emiss = (emiss * skvfac) + (1.0 - skvfac)

    # check for emissivity > 1.0
    emiss[emiss > 1.0] = 1.0

    # calculate incoming lw rad
    return emiss * STEF_BOLTZ * np.power(ta, 4)



import requests


class NWRCCheck(object):
    MYSQL_OPTIONS = {
        'user': 'unittest_user',
        'password': 'WsyR4Gp9JlFee6HwOHAQ',
        'host': '10.200.28.137',
        'database': 'weather_db',
        'metadata': 'tbl_metadata',
        'data_table': 'tbl_level2',
        'station_table': 'tbl_stations',
        'air_temp': 'air_temp',
        'vapor_pressure': 'vapor_pressure',
        'precip': 'precip_accum',
        'solar': 'solar_radiation',
        'wind_speed': 'wind_speed',
        'wind_direction': 'wind_direction',
        'cloud_factor': 'cloud_factor',
        'port': '32768'
    }
    URL = 'http://' + MYSQL_OPTIONS['host']

    @classmethod
    def in_network(cls):
        """
        Checks that were on the NWRC network
        """
        try:
            response = requests.get(cls.URL, timeout=1)
            return response.ok

        except requests.exceptions.RequestException:
            return False



from smrf.envphys import Snow
from smrf.envphys.nasde_model import Marks2017, PiecewiseSusong1999, Susong1999
import mock
import unittest
import numpy as np


class TestSnow(unittest.TestCase):
    TEMPERATURES = np.random.uniform(low=-8, high=3, size=(12, 12))
    PRECIPITATION = np.random.uniform(low=0, high=3, size=TEMPERATURES.shape)

    def test_missing_model(self):
        with self.assertRaises(ValueError):
            Snow.phase_and_density(self.TEMPERATURES, self.PRECIPITATION, '')

    @mock.patch.object(Susong1999, 'run')
    def test_passes_arguments(self, model_patch):
        Snow.phase_and_density(
            self.TEMPERATURES, self.PRECIPITATION, 'susong1999'
        )
        call_args = model_patch.call_args
        np.testing.assert_equal(self.TEMPERATURES, call_args[0][0])
        np.testing.assert_equal(self.PRECIPITATION, call_args[0][1])

    @mock.patch.object(Susong1999, 'run')
    def test_susong_1999(self, model_patch):
        Snow.phase_and_density(
            self.TEMPERATURES, self.PRECIPITATION, 'susong1999'
        )
        model_patch.assert_called()

    @mock.patch.object(PiecewiseSusong1999, 'run')
    def test_piecewise_susong_1999(self, model_patch):
        Snow.phase_and_density(
            self.TEMPERATURES, self.PRECIPITATION, 'piecewise_susong1999'
        )
        model_patch.assert_called()

    @mock.patch.object(Marks2017, 'run')
    def test_marks_2017(self, model_patch):
        Snow.phase_and_density(
            self.TEMPERATURES, self.PRECIPITATION, 'marks2017'
        )
        model_patch.assert_called()



import logging
import os
from datetime import datetime

import netCDF4 as nc
import numpy as np
from spatialnc.proj import add_proj, add_proj_from_file

from smrf import __version__


class OutputNetcdf():
    """
    Class OutputNetcdf() to output values to a netCDF file
    """

    type = 'netcdf'
    fmt = '%Y-%m-%d %H:%M:%S'
    cs = (6, 10, 10)

    def __init__(self, variable_dict, topo, time, outConfig):
        """
        Initialize the OutputNetcdf() class

        Args:
            variable_dict: dict of variable information
            topo: loadTopo instance
            time: time configuration
            outConfig: out configuration
        """

        self._logger = logging.getLogger(__name__)

        # go through the variable list and make full file names
        for v in variable_dict:
            variable_dict[v]['file_name'] = \
                str(variable_dict[v]['out_location'] + '.nc')

        self.variable_dict = variable_dict

        # process the time section
        self.run_time_step = int(time['time_step'])
        self.out_frequency = int(outConfig['frequency'])
        self.outConfig = outConfig

        # determine the x,y vectors for the netCDF file
        x = topo.x
        y = topo.y
        self.mask = topo.mask

        dimensions = ('time', 'y', 'x')
        self.date_time = {}
        now_str = datetime.now().strftime(self.fmt)

        # Retrieve projection information from topo
        map_meta = add_proj_from_file(topo.topoConfig['filename'])

        precision = self.outConfig['netcdf_output_precision'][0]

        for v in self.variable_dict:

            f = self.variable_dict[v]

            if os.path.isfile(f['file_name']):
                self._logger.warning('Opening {}, data may be overwritten!'
                                     .format(f['file_name']))

                # open in append mode
                s = nc.Dataset(f['file_name'], 'a')
                h = '[{}] Data added or updated'.format(now_str)
                setattr(s, 'last_modified', h)

                if 'projection' not in s.variables.keys():
                    s = add_proj(s, map_meta=map_meta)

            else:
                self._logger.debug('Creating %s' % f['file_name'])
                s = nc.Dataset(f['file_name'], 'w', format='NETCDF4',
                               clobber=True)

                # add dimensions
                s.createDimension(dimensions[0], None)
                s.createDimension(dimensions[1], y.shape[0])
                s.createDimension(dimensions[2], x.shape[0])

                # create the variables
                s.createVariable('time', 'f', (dimensions[0]))
                s.createVariable('y', 'f', dimensions[1])
                s.createVariable('x', 'f', dimensions[2])
                s.createVariable(f['variable'],
                                 precision,
                                 (dimensions[0], dimensions[1], dimensions[2]),
                                 chunksizes=self.cs)

                # # define some attributes
                s.variables['time'].setncattr(
                    'units',
                    'hours since {}'.format(time['start_date']))
                s.variables['time'].setncattr('calendar', 'standard')
                s.variables['time'].setncattr('time_zone', time['time_zone'])
                s.variables['time'].setncattr('long_name', 'time')

                # the y variable attributes
                s.variables['y'].setncattr('units', 'meters')
                s.variables['y'].setncattr('description', 'UTM, north south')
                s.variables['y'].setncattr('long_name', 'y coordinate')

                # the x variable attributes
                s.variables['x'].setncattr('units', 'meters')
                s.variables['x'].setncattr('description', 'UTM, east west')
                s.variables['x'].setncattr('long_name', 'x coordinate')

                # the variable attributes
                s.variables[f['variable']].setncattr(
                    'module',
                    f['module'])
                s.variables[f['variable']].setncattr(
                    'units',
                    f['info']['units'])
                s.variables[f['variable']].setncattr(
                    'long_name',
                    f['info']['long_name'])

                # define some global attribute
                s.setncattr_string('Conventions', 'CF-1.6')
                s.setncattr_string('dateCreated', now_str)
                s.setncattr_string('title', 'Distributed {0} data from SMRF'
                                            ''.format(f['info']['long_name']))
                s.setncattr_string('history', ('[{}] Create netCDF4 file'
                                               '').format(now_str))
                s.setncattr_string('institution', ('USDA Agricultural Research'
                                                   ' Service, Northwest'
                                                   ' Watershed Research'
                                                   ' Center'))
                s = add_proj(s, map_meta=map_meta)

                s.setncattr_string(
                    'references',
                    'Online documentation smrf.readthedocs.io;'
                    ' https://doi.org/10.1016/j.cageo.2017.08.016')

                s.variables['y'][:] = y
                s.variables['x'][:] = x

            s.setncattr_string('SMRF_version', __version__)
            s.close()

    def output(self, variable, data, date_time):
        """
        Output a time step

        Args:
            variable: variable name that will index into variable list
            data: the variable data
            date_time: the date time object for the time step
        """

        self._logger.debug('{0} Writing variable {1} to netCDF'
                           .format(date_time, variable))

#         f = self.variable_dict[variable]['nc_file']
        f = nc.Dataset(self.variable_dict[variable]['file_name'],
                       'a', 'NETCDF4')

        # the current time integer
        times = f.variables['time']
        t = nc.date2num(date_time.replace(tzinfo=None),
                        times.units,
                        times.calendar)

        if len(times) != 0:
            index = np.where(times[:] == t)[0]
            if index.size == 0:
                index = len(times)
            else:
                index = index[0]
        else:
            index = len(times)

        # insert the time and data
        f.variables['time'][index] = t

        if self.outConfig['mask_output']:
            f.variables[variable][index, :] = data*self.mask
        else:
            f.variables[variable][index, :] = data

        f.close()



import logging
import os

import numpy as np
import pytz
from scipy import interpolate

from smrf.distribute import image_data
from smrf.utils import utils


def interpx(yi, xi, x):
    """Interpolate in on direction

    Arguments:
        yi {array} -- y data to fit
        xi {array} -- x data to fit
        x {array} -- x data to interpolate over

    Returns:
        array -- y values evaluated at x
    """
    s = interpolate.interp1d(xi, yi, fill_value='extrapolate')
    return s(x)


class WindNinjaModel(image_data.image_data):
    """The `WindNinjaModel` loads data from a WindNinja simulation.
    The WindNinja is ran externally to SMRF and the configuration
    points to the location of the output ascii files. SMRF takes the
    files and interpolates to the model domain.
    """

    VARIABLE = 'wind'
    WN_DATE_FORMAT = '%m-%d-%Y_%H%M'
    DATE_FORMAT = '%Y%m%d'

    def __init__(self, smrf_config):
        """Initialize the WinstralWindModel

        Arguments:
            smrf_config {UserConfig.Dictionary} -- entire smrf config
            distribute_drifts {bool} -- not used

        Raises:
            IOError: if maxus file does not match topo size
        """

        image_data.image_data.__init__(self, self.VARIABLE)

        self._logger = logging.getLogger(__name__)
        self._logger.debug('Creating the WindNinjaModel')

        self.smrf_config = smrf_config
        self.getConfig(smrf_config['wind'])

        # wind ninja parameters
        self.wind_ninja_dir = self.config['wind_ninja_dir']
        self.wind_ninja_dxy = self.config['wind_ninja_dxdy']
        self.wind_ninja_pref = self.config['wind_ninja_pref']
        if self.config['wind_ninja_tz'] is not None:
            self.wind_ninja_tz = pytz.timezone(
                self.config['wind_ninja_tz'].title())

        # self.start_date = pd.to_datetime(
        #     self.smrf_config['time']['start_date'])
        # self.grid_data = self.smrf_config['gridded']['data_type']

        self.init_interp = True
        self.flatwind = None
        self.dir_round_cell = None
        self.cellmaxus = None

    def wind_ninja_path(self, dt, file_type):
        """Generate the path to the wind ninja data and ensure
        it exists.

        Arguments:
            file_type {str} -- type of file to get
        """

        # convert the SMRF date time to the WindNinja time
        t_file = dt.astimezone(self.wind_ninja_tz)

        f_path = os.path.join(
            self.wind_ninja_dir,
            'data{}'.format(dt.strftime(self.DATE_FORMAT)),
            'wind_ninja_data',
            '{}_{}_{:d}m_{}.asc'.format(
                self.wind_ninja_pref,
                t_file.strftime(self.WN_DATE_FORMAT),
                self.wind_ninja_dxy,
                file_type
            ))

        if not os.path.isfile(f_path):
            raise ValueError(
                'WindNinja file does not exist: {}!'.format(f_path))

        return f_path

    def initialize(self, topo, data=None):
        """Initialize the model with data

        Arguments:
            topo {topo class} -- Topo class
            data {None} -- Not used but needs to be there
        """

        # meshgrid points
        self.X = topo.X
        self.Y = topo.Y

        self.model_dxdy = np.mean(np.diff(topo.x))

        # WindNinja output height in meters
        self.wind_height = float(self.config['wind_ninja_height'])

        # set roughness that was used in WindNinja simulation
        # WindNinja uses 0.01m for grass, 0.43 for shrubs, and 1.0 for forest
        self.wn_roughness = float(self.config['wind_ninja_roughness']) * \
            np.ones_like(topo.dem)

        # get our effective veg surface roughness
        # to use in log law scaling of WindNinja data
        # using the relationship in
        # https://www.jstage.jst.go.jp/article/jmsj1965/53/1/53_1_96/_pdf
        self.veg_roughness = topo.veg_height / 7.39

        # make sure roughness stays reasonable using bounds from
        # http://www.iawe.org/Proceedings/11ACWE/11ACWE-Cataldo3.pdf
        self.veg_roughness[self.veg_roughness < 0.01] = 0.01
        self.veg_roughness[np.isnan(self.veg_roughness)] = 0.01
        self.veg_roughness[self.veg_roughness > 1.6] = 1.6

        # precalculate scale arrays so we don't do it every timestep
        self.ln_wind_scale = np.log(
            (self.veg_roughness + self.wind_height) / self.veg_roughness
        ) / np.log(
            (self.wn_roughness + self.wind_height) / self.wn_roughness
        )

    def initialize_interp(self, t):
        """Initialize the interpolation weights

        Arguments:
            t {datetime} -- initialize with this file
        """

        # do this first to speedup the interpolation later
        # find vertices and weights to speedup interpolation fro ascii file
        fp_vel = self.wind_ninja_path(t, 'vel')

        # get wind ninja topo stats
        ts2 = utils.get_asc_stats(fp_vel)
        self.windninja_x = ts2['x'][:]
        self.windninja_y = ts2['y'][:]

        XW, YW = np.meshgrid(self.windninja_x, self.windninja_y)
        self.wn_mx = XW.flatten()
        self.wn_my = YW.flatten()

        xy = np.zeros([XW.shape[0]*XW.shape[1], 2])
        xy[:, 1] = self.wn_my
        xy[:, 0] = self.wn_mx
        uv = np.zeros([self.X.shape[0]*self.X.shape[1], 2])
        uv[:, 1] = self.Y.flatten()
        uv[:, 0] = self.X.flatten()

        self.vtx, self.wts = utils.interp_weights(xy, uv, d=2)

        self.init_interp = False

    def distribute(self, data_speed, data_direction):
        """Distribute the wind for the model

        Arguments:
            data_speed {DataFrame} -- wind speed data frame
            data_direction {DataFrame} -- wind direction data frame
        """

        t = data_speed.name

        if self.init_interp:
            self.initialize_interp(t)

        wind_speed, wind_direction = self.convert_wind_ninja(t)
        self.wind_speed = wind_speed
        self.wind_direction = wind_direction

    def convert_wind_ninja(self, t):
        """
        Convert the WindNinja ascii grids back to the SMRF grids and into the
        SMRF data streamself.

        Args:
            t: datetime of timestep

        Returns:
            ws: wind speed numpy array
            wd: wind direction numpy array

        """

        # get the ascii files that need converted
        fp_vel = self.wind_ninja_path(t, 'vel')
        data_vel = np.loadtxt(fp_vel, skiprows=6)
        data_vel_int = data_vel.flatten()

        # interpolate to the SMRF grid from the WindNinja grid
        g_vel = utils.grid_interpolate(
            data_vel_int, self.vtx,
            self.wts, self.X.shape)

        # There will be NaN's around the edge, hanlde those first
        if self.model_dxdy != self.wind_ninja_dxy:
            self._logger.debug('Wind speed from WindNinja has NaN, filling')

            g_vel = self.fill_data(g_vel)

        # log law scale
        g_vel = g_vel * self.ln_wind_scale

        # wind direction from angle, split into u,v components then interpolate
        fp_ang = self.wind_ninja_path(t, 'ang')
        data_ang = np.loadtxt(fp_ang, skiprows=6)

        u = np.sin(data_ang * np.pi / 180)
        v = np.cos(data_ang * np.pi / 180)

        ui = utils.grid_interpolate(
            u.flatten(), self.vtx, self.wts, self.X.shape)
        vi = utils.grid_interpolate(
            v.flatten(), self.vtx, self.wts, self.X.shape)

        uf = self.fill_data(ui)
        vf = self.fill_data(vi)

        g_ang = np.arctan2(uf, vf) * 180 / np.pi
        g_ang[g_ang < 0] = g_ang[g_ang < 0] + 360

        return g_vel, g_ang

    def fill_data(self, g_vel):
        """Fill the WindNinja array that has NaN's.
        This makes an assumption that all the NaN values are along
        the left and bottom edge. This will be the case in the Northern
        hemisphere. First fill the Y direction with 1d interpolation
        exprapolated to the edges, then do the same in the X direction.
        At the end, it will check to ensure that there are no NaN values
        left.

        Arguments:
            g_vel {np.array} -- numpy array to fill

        Raises:
            ValueError: If there are still NaN values after filling

        Returns:
            np.array -- filled numpy array
        """

        ix = np.sum(np.isnan(g_vel[0, :]))
        iy = np.sum(np.isnan(g_vel[:, ix+1]))

        #  first go in the Y direction
        yi = g_vel[:, ix:ix+10]
        xi = self.X[0, ix:ix+10]
        x = self.X[0, :ix]

        o = np.apply_along_axis(interpx, axis=1, arr=yi, xi=xi, x=x)
        g_vel[:, :ix] = o

        #  first go in the X direction
        yi = g_vel[-iy-10:-iy, :]
        xi = self.Y[-iy-10:-iy, 0]
        x = self.Y[-iy:, 0]

        o = np.apply_along_axis(interpx, axis=0, arr=yi, xi=xi, x=x)
        g_vel[-iy:, :] = o

        if np.any(np.isnan(g_vel)):
            raise ValueError('WindNinja data still has NaN values')

        return g_vel



# -*- coding: utf-8 -*-
# flake8: noqa

import os

from pkg_resources import DistributionNotFound, get_distribution

try:
    __version__ = get_distribution(__name__).version
except DistributionNotFound:
    __version__ = 'unknown'

__core_config__ = os.path.abspath(
    os.path.dirname(__file__) + '/framework/CoreConfig.ini')
__recipes__ = os.path.abspath(os.path.dirname(
    __file__) + '/framework/recipes.ini')
__config_changelog__ = os.path.abspath(
    os.path.dirname(__file__) + '/framework/changelog.ini')

__config_titles__ = {
    "topo": "Files for DEM and vegetation",
    "time": "Dates to run model",
    "stations": "Stations to use",
    "csv": "CSV section configurations",
    "gridded": "Gridded datasets configurations",
    "air_temp": "Air temperature distribution",
    "vapor_pressure": "Vapor pressure distribution",
    "wind": "Wind speed and wind direction distribution",
    "precip": "Precipitation distribution",
    "albedo": "Albedo distribution",
    "cloud_factor": "Cloud Factor - Fraction used to limit solar radiation Cloudy (0) - Sunny (1)",
    "solar": "Solar radiation",
    "thermal": "Incoming thermal radiation",
    "soil_temp": "Soil temperature",
    "output": "Output variables",
    "system": "System variables and Logging"
}

from . import utils, data, distribute, envphys, framework, output, spatial  # isort:skip

__config_header__ = "Config File for SMRF {0}\n" \
                    "For more SMRF related help see:\n" \
                    "{1}".format(
                        __version__, 'http://smrf.readthedocs.io/en/latest/')
__config_checkers__ = 'utils.utils'
# -*- coding: utf-8 -*-



"""
The module :mod:`~smrf.framework.model_framework` contains functions and
classes that act as a major wrapper to the underlying packages and modules
contained with SMRF. A class instance of
:class:`~smrf.framework.model_framework.SMRF` is initialized with a
configuration file indicating where data is located, what variables to
distribute and how, where to output the distributed data, or run as a threaded
application. See the help on the configuration file to learn more about how to
control the actions of :class:`~smrf.framework.model_framework.SMRF`.

Example:
    The following examples shows the most generic method of running SMRF. These
    commands will generate all the forcing data required to run iSnobal.  A
    complete example can be found in run_smrf.py

    >>> import smrf
    >>> s = smrf.framework.SMRF(configFile) # initialize SMRF
    >>> s.loadTopo() # load topo data
    >>> s.create_distribution() # initialize the distribution
    >>> s.initializeOutput() # initialize the outputs if desired
    >>> s.loadData() # load weather data  and station metadata
    >>> s.disttribute_data() # distribute

"""

import logging
import os
import sys
from datetime import datetime
from os.path import abspath, join
from threading import Thread

import numpy as np
import pandas as pd
import pytz
from inicheck.config import UserConfig
from inicheck.output import generate_config, print_config_report
from inicheck.tools import check_config, get_user_config
from topocalc.shade import shade

from smrf import distribute
from smrf.data import InputData, Topo
from smrf.envphys import sunang
from smrf.envphys.solar import model
from smrf.framework import art, logger
from smrf.output import output_hru, output_netcdf
from smrf.utils import queue
from smrf.utils.utils import backup_input, date_range, getqotw


class SMRF():
    """
    SMRF - Spatial Modeling for Resources Framework

    Args:
        configFile (str):  path to configuration file.

    Returns:
        SMRF class instance.

    Attributes:
        start_date: start_date read from configFile
        end_date: end_date read from configFile
        date_time: Numpy array of date_time objects between start_date and
            end_date
        config: Configuration file read in as dictionary
        distribute: Dictionary the contains all the desired variables to
            distribute and is initialized in
            :func:`~smrf.framework.model_framework.create_distribution`
    """

    # These are the modules that the user can modify and use different methods
    modules = ['air_temp',
               'albedo',
               'precip',
               'soil_temp',
               'solar',
               'cloud_factor',
               'thermal',
               'vapor_pressure',
               'wind']

    BASE_THREAD_VARIABLES = frozenset([
        'cosz', 'azimuth', 'illum_ang', 'output'
    ])

    def __init__(self, config, external_logger=None):
        """
        Initialize the model, read config file, start and end date, and logging
        """
        # read the config file and store
        if isinstance(config, str):
            if not os.path.isfile(config):
                raise Exception('Configuration file does not exist --> {}'
                                .format(config))
            self.configFile = config

            # Read in the original users config
            ucfg = get_user_config(config, modules='smrf')

        elif isinstance(config, UserConfig):
            ucfg = config
            self.configFile = config.filename

        else:
            raise Exception('Config passed to SMRF is neither file name nor '
                            ' UserConfig instance')
        # start logging
        if external_logger is None:
            self.smrf_logger = logger.SMRFLogger(ucfg.cfg['system'])
            self._logger = logging.getLogger(__name__)
        else:
            self._logger = external_logger

        # add the title
        self.title(2)

        # Make the output directory if it do not exist
        out = ucfg.cfg['output']['out_location']
        os.makedirs(out, exist_ok=True)

        # Check the user config file for errors and report issues if any
        self._logger.info("Checking config file for issues...")
        warnings, errors = check_config(ucfg)
        print_config_report(warnings, errors, logger=self._logger)
        self.ucfg = ucfg
        self.config = self.ucfg.cfg

        # Exit SMRF if config file has errors
        if len(errors) > 0:
            self._logger.error("Errors in the config file. See configuration"
                               " status report above.")
            sys.exit()

        # Write the config file to the output dir
        full_config_out = abspath(join(out, 'config.ini'))

        self._logger.info("Writing config file with full options.")
        generate_config(self.ucfg, full_config_out)

        # Process the system variables
        for k, v in self.config['system'].items():
            setattr(self, k, v)

        self._setup_date_and_time()

        # need to align date time
        if 'date_method_start_decay' in self.config['albedo'].keys():
            self.config['albedo']['date_method_start_decay'] = \
                self.config['albedo']['date_method_start_decay'].replace(
                    tzinfo=self.time_zone)
            self.config['albedo']['date_method_end_decay'] = \
                self.config['albedo']['date_method_end_decay'].replace(
                    tzinfo=self.time_zone)

        # if a gridded dataset will be used
        self.gridded = False
        self.forecast_flag = False
        self.hrrr_data_timestep = False
        if 'gridded' in self.config:
            self.gridded = True
            if self.config['gridded']['data_type'] in ['hrrr_grib']:
                self.forecast_flag = \
                    self.config['gridded']['hrrr_forecast_flag']
                self.hrrr_data_timestep = \
                    self.config['gridded']['hrrr_load_method'] == 'timestep'

        now = datetime.now().astimezone(self.time_zone)
        if ((self.start_date > now and not self.gridded) or
                (self.end_date > now and not self.gridded)):
            raise ValueError("A date set in the future can only be used with"
                             " WRF generated data!")

        self.distribute = {}

        if self.config['system']['qotw']:
            self._logger.info(getqotw())

        # Initialize the distribute dict
        self._logger.info('Started SMRF --> %s' % now)
        self._logger.info('Model start --> %s' % self.start_date)
        self._logger.info('Model end --> %s' % self.end_date)
        self._logger.info('Number of time steps --> %i' % self.time_steps)

    def _setup_date_and_time(self):
        self.time_zone = pytz.timezone(self.config['time']['time_zone'])
        is_utz = self.time_zone == pytz.UTC

        # Get the time section utils
        self.start_date = pd.to_datetime(
            self.config['time']['start_date'], utc=is_utz
        )
        self.end_date = pd.to_datetime(
            self.config['time']['end_date'], utc=is_utz
        )

        if not is_utz:
            self.start_date = self.start_date.tz_localize(self.time_zone)
            self.end_date = self.end_date.tz_localize(self.time_zone)

        # Get the time steps correctly in the time zone
        self.date_time = date_range(
            self.start_date,
            self.end_date,
            self.config['time']['time_step'],
            self.time_zone
        )
        self.time_steps = len(self.date_time)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        """
        Provide some logging info about when SMRF was closed
        """

        self._logger.info('SMRF closed --> %s' % datetime.now())
        logging.shutdown()

    @property
    def possible_output_variables(self):
        # Collect the potential output variables
        variables = {}
        for variable, module in self.distribute.items():
            variables.update(module.output_variables)
        return variables

    def loadTopo(self):
        """
        Load the information from the configFile in the ['topo'] section. See
        :func:`smrf.data.loadTopo.Topo` for full description.
        """

        self.topo = Topo(self.config['topo'])

    def create_distribution(self):
        """
        This initializes the distirbution classes based on the configFile
        sections for each variable.
        :func:`~smrf.framework.model_framework.SMRF.create_distribution`
        will initialize the variables within the :func:`smrf.distribute`
        package and insert into a dictionary 'distribute' with variable names
        as the keys.

        Variables that are intialized are:
            * :func:`Air temperature <smrf.distribute.air_temp.ta>`
            * :func:`Vapor pressure <smrf.distribute.vapor_pressure.vp>`
            * :func:`Wind speed and direction <smrf.distribute.wind.wind>`
            * :func:`Precipitation <smrf.distribute.precipitation.ppt>`
            * :func:`Albedo <smrf.distribute.albedo.Albedo>`
            * :func:`Solar radiation <smrf.distribute.solar.Solar>`
            * :func:`Thermal radiation <smrf.distribute.thermal.th>`
            * :func:`Soil Temperature <smrf.distribute.soil_temp.ts>`
        """

        # 1. Air temperature
        self.distribute['air_temp'] = distribute.air_temp.ta(
            self.config['air_temp'])

        # 2. Vapor pressure
        self.distribute['vapor_pressure'] = distribute.vapor_pressure.vp(
            self.config['vapor_pressure'],
            self.config['precip']['precip_temp_method'])

        # 3. Wind
        self.distribute['wind'] = distribute.wind.Wind(self.config)

        # 4. Precipitation
        self.distribute['precipitation'] = distribute.precipitation.ppt(
            self.config['precip'],
            self.start_date,
            self.config['time']['time_step'])

        # 5. Albedo
        self.distribute['albedo'] = distribute.albedo.Albedo(
            self.config['albedo'])

        # 6. cloud_factor
        self.distribute['cloud_factor'] = distribute.cloud_factor.cf(
            self.config['cloud_factor'])

        # 7. Solar radiation
        self.distribute['solar'] = distribute.solar.Solar(
            self.config,
            self.topo)

        # 8. thermal radiation
        self.distribute['thermal'] = distribute.thermal.th(
            self.config['thermal'])

        # 9. soil temperature
        self.distribute['soil_temp'] = distribute.soil_temp.ts(
            self.config['soil_temp'])

    def loadData(self):
        """
        Load the measurement point data for distributing to the DEM,
        must be called after the distributions are initialized. Currently, data
        can be loaded from three different sources:

            * :func:`CSV files <smrf.data.loadData.wxdata>`
            * :func:`MySQL database <smrf.data.loadData.wxdata>`
            * :func:`Gridded data source (WRF) <smrf.data.loadGrid.grid>`

        After loading, :func:`~smrf.framework.mode_framework.SMRF.loadData`
        will call :func:`smrf.framework.model_framework.find_pixel_location`
        to determine the pixel locations of the point measurements and filter
        the data to the desired stations if CSV files are used.
        """

        self.data = InputData(
            self.config,
            self.start_date,
            self.end_date,
            self.topo)

        # Pre-filter the data to the desired stations in
        # each [variable] section
        self._logger.debug(
            'Filter data to those specified in each variable section')
        for variable, module in self.data.MODULE_VARIABLES.items():

            # Check to find the matching stations
            data = getattr(self.data, variable, pd.DataFrame())
            if self.distribute[module].stations is not None:

                match = data.columns.isin(self.distribute[module].stations)
                sta_match = data.columns[match]

                # Update the dataframe and the distribute stations
                self.distribute[module].stations = sta_match.tolist()
                setattr(self.data, variable, data[sta_match])

            else:
                self.distribute[module].stations = data.columns.tolist()

        # Does the user want to create a CSV copy of the station data used.
        if self.config["output"]['input_backup']:
            self._logger.info('Backing up input data...')
            backup_input(self.data, self.ucfg)

    def disttribute_data(self):
        """
        Wrapper for various distribute methods. If threading was set in
        configFile, then
        :func:`~smrf.framework.model_framework.SMRF.disttribute_data_threaded`
        will be called. Default will call
        :func:`~smrf.framework.model_framework.SMRF.disttribute_data_serial`.
        """

        if self.threading:
            self.disttribute_data_threaded()
        else:
            self.disttribute_data_serial()

    def initialize_distribution(self, date_time=None):
        """Call the initialize method for each distribute module

        Args:
            date_time (list, optional): initialize with the datetime list
                or not. Defaults to None.
        """

        for v in self.distribute:
            self.distribute[v].initialize(self.topo, self.data, date_time)

    def disttribute_data_serial(self):
        """
        Distribute the measurement point data for all variables in serial. Each
        variable is initialized first using the :func:`smrf.data.loadTopo.Topo`
        instance and the metadata loaded from
        :func:`~smrf.framework.model_framework.SMRF.loadData`.
        The function distributes over each time step, all the variables below.

        Steps performed:
            1. Sun angle for the time step
            2. Illumination angle
            3. Air temperature
            4. Vapor pressure
            5. Wind direction and speed
            6. Precipitation
            7. Cloud Factor
            8. Solar radiation
            9. Thermal radiation
            10. Soil temperature
            11. Output time step if needed
        """

        self.initialize_distribution()

        # -------------------------------------
        # Distribute the data
        for output_count, t in enumerate(self.date_time):

            startTime = datetime.now()

            self.distribute_single_timestep(t)
            self.output(t)

            telapsed = datetime.now() - startTime
            self._logger.debug('{0:.2f} seconds for time step'
                               .format(telapsed.total_seconds()))

        self.forcing_data = 1

    def distribute_single_timestep(self, t):

        self._logger.info('Distributing time step {}'.format(t))

        if self.hrrr_data_timestep:
            self.data.load_class.load_timestep(t)
            self.data.set_variables()

        # 0.1 sun angle for time step
        cosz, azimuth, rad_vec = sunang.sunang(
            t.astimezone(pytz.utc),
            self.topo.basin_lat,
            self.topo.basin_long)

        # 0.2 illumination angle
        illum_ang = None
        if cosz > 0:
            illum_ang = shade(
                self.topo.sin_slope,
                self.topo.aspect,
                azimuth,
                cosz)

        # 1. Air temperature
        self.distribute['air_temp'].distribute(self.data.air_temp.loc[t])

        # 2. Vapor pressure
        self.distribute['vapor_pressure'].distribute(
            self.data.vapor_pressure.loc[t],
            self.distribute['air_temp'].air_temp)

        # 3. Wind_speed and wind_direction
        self.distribute['wind'].distribute(
            self.data.wind_speed.loc[t],
            self.data.wind_direction.loc[t],
            t)

        # 4. Precipitation
        self.distribute['precipitation'].distribute(
            self.data.precip.loc[t],
            self.distribute['vapor_pressure'].dew_point,
            self.distribute['vapor_pressure'].precip_temp,
            self.distribute['air_temp'].air_temp,
            t,
            self.data.wind_speed.loc[t],
            self.data.air_temp.loc[t],
            self.distribute['wind'].wind_direction,
            self.distribute['wind'].wind_model.dir_round_cell,
            self.distribute['wind'].wind_speed,
            self.distribute['wind'].wind_model.cellmaxus)

        # 5. Albedo
        self.distribute['albedo'].distribute(
            t,
            illum_ang,
            self.distribute['precipitation'].storm_days)

        # 6. cloud_factor
        self.distribute['cloud_factor'].distribute(
            self.data.cloud_factor.loc[t])

        # 7. Solar
        self.distribute['solar'].distribute(
            t,
            self.distribute["cloud_factor"].cloud_factor,
            illum_ang,
            cosz,
            azimuth,
            self.distribute['albedo'].albedo_vis,
            self.distribute['albedo'].albedo_ir)

        # 8. thermal radiation
        self.distribute['thermal'].distribute(
            t,
            self.distribute['air_temp'].air_temp,
            self.distribute['vapor_pressure'].vapor_pressure,
            self.distribute['vapor_pressure'].dew_point,
            self.distribute['cloud_factor'].cloud_factor)

        # 9. Soil temperature
        self.distribute['soil_temp'].distribute()

    def disttribute_data_threaded(self):
        """
        Distribute the measurement point data for all variables using threading
        and queues. Each variable is initialized first using the
        :func:`smrf.data.loadTopo.Topo` instance and the metadata loaded from
        :func:`~smrf.framework.model_framework.SMRF.loadData`. A
        :func:`DateQueue <smrf.utils.queue.DateQueue_Threading>` is initialized
        for :attr:`all threading
        variables <smrf.framework.model_framework.SMRF.thread_variables>`. Each
        variable in :func:`smrf.distribute` is passed all the required point
        data at once using the distribute_thread function.  The
        distribute_thread function iterates over
        :attr:`~smrf.framework.model_framework.SMRF.date_time` and places the
        distributed values into the
        :func:`DateQueue <smrf.utils.queue.DateQueue_Threading>`.
        """

        # Load the data into the data queue
        self.create_data_queue()

        # Create threads for distribution
        self.create_distributed_threads()

        # output thread
        self.threads.append(
            queue.QueueOutput(
                self.smrf_queue,
                self.date_time,
                self.out_func,
                self.config['output']['frequency'],
                self.topo.nx,
                self.topo.ny))

        # the cleaner
        self.threads.append(queue.QueueCleaner(
            self.date_time, self.smrf_queue))

        # start all the threads
        for i in range(len(self.threads)):
            self.threads[i].start()

        # Wait for the end
        for i in range(len(self.threads)):
            self.threads[i].join()

        self._logger.debug('DONE!!!!')

    def create_data_queue(self):

        self._logger.info('Creating the data queue and loading current data')

        self.data_queue = {}
        for variable in self.data.VARIABLES[:-1]:
            dq = queue.DateQueueThreading(
                timeout=self.time_out,
                name="data_{}".format(variable))

            # load the data into the queue, all methods should have
            # loaded something, even the HRRR will have a single hour
            # of data loaded.
            data = getattr(self.data, variable, pd.DataFrame())
            for date_time, row in data.iterrows():
                dq.put([date_time, row])

            self.data_queue[variable] = dq

        # create a thread to load the data
        if self.hrrr_data_timestep:
            data_thread = Thread(
                target=self.data.load_class.load_timestep_thread,
                name='data',
                args=(self.date_time, self.data_queue))
            data_thread.start()

    def set_queue_variables(self):

        # These are the variables that will be queued
        self.thread_queue_variables = list(self.BASE_THREAD_VARIABLES)

        for v in self.distribute:
            self.thread_queue_variables += self.distribute[v].thread_variables

    def create_distributed_threads(self, other_queue=None):
        """
        Creates the threads for a distributed run in smrf.
        Designed for smrf runs in memory

        Returns
            t: list of threads for distirbution
            q: queue
        """

        # -------------------------------------
        # Initialize the distributions and get thread variables
        self._logger.info("Initializing distributed variables...")

        self.initialize_distribution(self.date_time)
        self.set_queue_variables()

        # -------------------------------------
        # Create Queues for all the variables
        self.smrf_queue = {}
        self._logger.info("Staging {} threaded variables...".format(
            len(self.thread_queue_variables)))
        for v in self.thread_queue_variables:
            self.smrf_queue[v] = queue.DateQueueThreading(
                self.queue_max_values,
                self.time_out,
                name=v)

        # -------------------------------------
        # Distribute the data
        self.threads = []

        # 0.1 sun angle for time step
        self.threads.append(Thread(
            target=sunang.sunang_thread,
            name='sun_angle',
            args=(self.smrf_queue, self.date_time,
                  self.topo.basin_lat,
                  self.topo.basin_long)))

        # 0.2 illumination angle
        self.threads.append(Thread(
            target=model.shade_thread,
            name='illum_angle',
            args=(self.smrf_queue, self.date_time,
                  self.topo.sin_slope, self.topo.aspect)))

        thread_dict = {
            'air_temp': self.distribute['air_temp'].distribute_thread,
            'vapor_pressure': self.distribute['vapor_pressure'].distribute_thread,  # noqa
            'wind': self.distribute['wind'].distribute_thread,
            'precipitation': self.distribute['precipitation'].distribute_thread,  # noqa
            'albedo': self.distribute['albedo'].distribute_thread,
            'cloud_factor': self.distribute['cloud_factor'].distribute_thread,
            'solar': self.distribute['solar'].distribute_thread,
            'thermal': self.distribute['thermal'].distribute_thread
        }

        for name, target in thread_dict.items():
            self.threads.append(
                Thread(
                    target=target,
                    name=name,
                    args=(self.smrf_queue, self.data_queue))
            )

    def create_output_variable_dict(self, output_variables, out_location):

        # determine which variables belong where
        variable_dict = {}
        for output_variable in output_variables:

            if output_variable in self.possible_output_variables.keys():
                fname = join(out_location, output_variable)
                module = self.possible_output_variables[output_variable]['module']  # noqa

                # TODO this is a hack to not have to redo the gold files
                if module == 'precipitation':
                    nc_module = 'precip'
                else:
                    nc_module = module

                variable_dict[output_variable] = {
                    'variable': output_variable,
                    'module': nc_module,
                    'out_location': fname,
                    'info': self.distribute[module].output_variables[output_variable]  # noqa
                }

            else:
                self._logger.error(
                    '{} not an output variable'.format(output_variable))

        return variable_dict

    def initializeOutput(self):
        """
        Initialize the output files based on the configFile section ['output'].
        Currently only :func:`NetCDF files
        <smrf.output.output_netcdf.OutputNetcdf>` are supported.
        """
        out_location = self.config['output']['out_location']

        # determine the variables to be output
        s_var_list = ", ".join(self.config['output']['variables'])
        self._logger.info('{} variables will be output'.format(s_var_list))

        output_variables = self.config['output']['variables']

        variable_dict = self.create_output_variable_dict(
            output_variables, out_location)

        self._logger.debug('{} of {} variables will be output'.format(
            len(output_variables), len(self.possible_output_variables)))

        # determine what type of file to output
        if self.config['output']['file_type'].lower() == 'netcdf':
            self.out_func = output_netcdf.OutputNetcdf(
                variable_dict, self.topo,
                self.config['time'],
                self.config['output'])

        elif self.config['output']['file_type'].lower() == 'hru':
            self.out_func = output_hru.output_hru(
                variable_dict, self.topo,
                self.date_time,
                self.config['output'])

        else:
            raise Exception('Could not determine type of file for output')

        # is there a function to apply?
        self.out_func.func = None
        if 'func' in self.config['output']:
            self.out_func.func = self.config['output']['func']

    def output(self, current_time_step,  module=None, out_var=None):
        """
        Output the forcing data or model outputs for the current_time_step.

        Args:
            current_time_step (date_time): the current time step datetime
                                            object

            module (str): module name
            out_var (str) - output a single variable

        """
        output_count = self.date_time.index(current_time_step)

        # Only output according to the user specified value,
        # or if it is the end.
        if (output_count % self.config['output']['frequency'] == 0) or \
           (output_count == len(self.date_time)):

            # User is attempting to output single variable
            if module is not None and out_var is not None:
                # add only one variable to the output list and preceed
                var_vals = [self.out_func.variable_dict[out_var]]

            # Incomplete request
            elif module is not None or out_var is not None:
                raise ValueError("Function requires an output module and"
                                 " variable name when outputting a specific"
                                 " variables")

            else:
                # Output all the variables
                var_vals = self.out_func.variable_dict.values()

            # Get the output variables then pass to the function
            for v in var_vals:
                # get the data desired
                data = getattr(
                    self.distribute[v['info']['module']], v['variable'])

                if data is None:
                    data = np.zeros((self.topo.ny, self.topo.nx))

                # output the time step
                self._logger.debug("Outputting {0}".format(v['module']))
                self.out_func.output(v['variable'], data, current_time_step)

    def post_process(self):
        """
        Execute all the post processors
        """

        for k in self.distribute.keys():
            self.distribute[k].post_processor(self)

    def title(self, option):
        """
        A little title to go at the top of the logger file
        """

        if option == 1:
            title = art.title1

        elif option == 2:
            title = art.title2

        for line in title:
            self._logger.info(line)


def run_smrf(config, external_logger=None):
    """
    Function that runs smrf how it should be operate for full runs.

    Args:
        config: string path to the config file or inicheck UserConfig instance
    """
    start = datetime.now()
    # initialize
    with SMRF(config, external_logger) as s:
        # load topo data
        s.loadTopo()

        # initialize the distribution
        s.create_distribution()

        # initialize the outputs if desired
        s.initializeOutput()

        # load weather data  and station metadata
        s.loadData()

        # distribute
        s.disttribute_data()

        # post process if necessary
        s.post_process()

        s._logger.info(datetime.now() - start)

    return s


def can_i_run_smrf(config):
    """
    Function that wraps run_smrf in try, except for testing purposes

    Args:
        config: string path to the config file or inicheck UserConfig instance
    """
    try:
        run_smrf(config)
        return True
    except Exception as e:
        raise e
        return False
